# -*- coding: utf-8 -*-
# Copyright 2007-2020 The HyperSpy developers
#
# This file is part of  HyperSpy.
#
#  HyperSpy is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
#  HyperSpy is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with  HyperSpy.  If not, see <http://www.gnu.org/licenses/>.

import copy
import itertools
from distutils.version import LooseVersion
import textwrap
import traits.api as t
import matplotlib.pyplot as plt
import matplotlib as mpl
from mpl_toolkits.axes_grid1 import make_axes_locatable
from matplotlib.backend_bases import key_press_handler
import warnings
import numpy as np
import logging
from functools import partial

import hyperspy as hs
from hyperspy.defaults_parser import preferences


_logger = logging.getLogger(__name__)


def contrast_stretching(data, vmin=None, vmax=None):
    """Estimate bounds of the data to display.

    Parameters
    ----------
    data: numpy array
    vmin, vmax: scalar, str, None
        If str, formatted as 'xth', use this value to calculate the percentage
        of pixels that are left out of the lower and upper bounds.
        For example, for a vmin of '1th', 1% of the lowest will be ignored to
        estimate the minimum value. Similarly, for a vmax value of '1th', 1%
        of the highest value will be ignored in the estimation of the maximum
        value. See :py:func:`numpy.percentile` for more explanation.
        If None, use the percentiles value set in the preferences.
        If float of integer, keep this value as bounds.

    Returns
    -------
    vmin, vmax: scalar
        The low and high bounds.

    Raises
    ------
    ValueError
        if the value of `vmin` `vmax` is out of the valid range for percentile
        calculation (in case of string values).

    """
    def _parse_value(value, value_name):
        if value is None:
            if value_name == "vmin":
                value = f'{preferences.Plot.saturated_pixels / 2}th'
            elif value_name == "vmax":
                value = f'{100 -  preferences.Plot.saturated_pixels / 2}th'
        if isinstance(value, str):
            value = float(value.split("th")[0])
        if not 0 <= value <= 100:
            raise ValueError(f"{value_name} must be in the range[0, 100].")
        return value

    if np.ma.is_masked(data):
        # If there is a mask, compressed the data to remove the masked data
        data = np.ma.masked_less_equal(data, 0).compressed()

    # If vmin, vmax are float or int, we keep the value, if not we calculate
    # the precentile value
    if not isinstance(vmin, (float, int)):
        vmin = np.nanpercentile(data, _parse_value(vmin, 'vmin'))
    if not isinstance(vmax, (float, int)):
        vmax = np.nanpercentile(data, _parse_value(vmax, 'vmax'))

    return vmin, vmax


MPL_DIVERGING_COLORMAPS = [
    "BrBG",
    "bwr",
    "coolwarm",
    "PiYG",
    "PRGn",
    "PuOr",
    "RdBu",
    "RdGy",
    "RdYIBu",
    "RdYIGn",
    "seismic",
    "Spectral", ]
# Add reversed colormaps
MPL_DIVERGING_COLORMAPS += [cmap + "_r" for cmap in MPL_DIVERGING_COLORMAPS]


def centre_colormap_values(vmin, vmax):
    """Calculate vmin and vmax to set the colormap midpoint to zero.

    Parameters
    ----------
    vmin, vmax : scalar
        The range of data to display.

    Returns
    -------
    cvmin, cvmax : scalar
        The values to obtain a centre colormap.

    """

    absmax = max(abs(vmin), abs(vmax))
    return -absmax, absmax


def create_figure(window_title=None,
                  _on_figure_window_close=None,
                  disable_xyscale_keys=False,
                  **kwargs):
    """Create a matplotlib figure.

    This function adds the possibility to execute another function
    when the figure is closed and to easily set the window title. Any
    keyword argument is passed to the plt.figure function.

    Parameters
    ----------
    window_title : {None, string}, optional
    _on_figure_window_close : {None, function}, optional
    disable_xyscale_keys : bool, optional
        Disable the `k`, `l` and `L` shortcuts which toggle the x or y axis
        between linear and log scale. Default False.

    Returns
    -------
    fig : plt.figure

    """
    fig = plt.figure(**kwargs)
    if window_title is not None:
        # remove non-alphanumeric characters to prevent file saving problems
        # This is a workaround for:
        #   https://github.com/matplotlib/matplotlib/issues/9056
        reserved_characters = r'<>"/\|?*'
        for c in reserved_characters:
            window_title = window_title.replace(c, '')
        window_title = window_title.replace('\n', ' ')
        window_title = window_title.replace(':', ' -')
        fig.canvas.set_window_title(window_title)
    if disable_xyscale_keys and hasattr(fig.canvas, 'toolbar'):
        # hack the `key_press_handler` to disable the `k`, `l`, `L` shortcuts
        manager = fig.canvas.manager
        fig.canvas.mpl_disconnect(manager.key_press_handler_id)
        manager.key_press_handler_id = manager.canvas.mpl_connect(
            'key_press_event',
            lambda event: key_press_handler_custom(event, manager.canvas))
    if _on_figure_window_close is not None:
        on_figure_window_close(fig, _on_figure_window_close)
    return fig


def key_press_handler_custom(event, canvas):
    if event.key not in ['k', 'l', 'L']:
        key_press_handler(event, canvas, canvas.manager.toolbar)


def on_figure_window_close(figure, function):
    """Connects a close figure signal to a given function.

    Parameters
    ----------

    figure : mpl figure instance
    function : function

    """
    def function_wrapper(evt):
        function()

    figure.canvas.mpl_connect('close_event', function_wrapper)


def plot_RGB_map(im_list, normalization='single', dont_plot=False):
    """Plot 2 or 3 maps in RGB.

    Parameters
    ----------
    im_list : list of Signal2D instances
    normalization : {'single', 'global'}, optional
    dont_plot : bool, optional
        Default False.

    Returns
    -------
    array: RGB matrix

    """
#    from widgets import cursors
    height, width = im_list[0].data.shape[:2]
    rgb = np.zeros((height, width, 3))
    rgb[:, :, 0] = im_list[0].data.squeeze()
    rgb[:, :, 1] = im_list[1].data.squeeze()
    if len(im_list) == 3:
        rgb[:, :, 2] = im_list[2].data.squeeze()
    if normalization == 'single':
        for i in range(len(im_list)):
            rgb[:, :, i] /= rgb[:, :, i].max()
    elif normalization == 'global':
        rgb /= rgb.max()
    rgb = rgb.clip(0, rgb.max())
    if not dont_plot:
        figure = plt.figure()
        ax = figure.add_subplot(111)
        ax.frameon = False
        ax.set_axis_off()
        ax.imshow(rgb, interpolation='nearest')
#        cursors.set_mpl_ax(ax)
        figure.canvas.draw_idle()
    else:
        return rgb


def subplot_parameters(fig):
    """Returns a list of the subplot parameters of a mpl figure.

    Parameters
    ----------
    fig : mpl figure

    Returns
    -------
    tuple : (left, bottom, right, top, wspace, hspace)

    """
    wspace = fig.subplotpars.wspace
    hspace = fig.subplotpars.hspace
    left = fig.subplotpars.left
    right = fig.subplotpars.right
    top = fig.subplotpars.top
    bottom = fig.subplotpars.bottom
    return left, bottom, right, top, wspace, hspace


class ColorCycle:
    _color_cycle = [mpl.colors.colorConverter.to_rgba(color) for color
                    in ('b', 'g', 'r', 'c', 'm', 'y', 'k')]

    def __init__(self):
        self.color_cycle = copy.copy(self._color_cycle)

    def __call__(self):
        if not self.color_cycle:
            self.color_cycle = copy.copy(self._color_cycle)
        return self.color_cycle.pop(0)


def plot_signals(signal_list, sync=True, navigator="auto",
                 navigator_list=None, **kwargs):
    """Plot several signals at the same time.

    Parameters
    ----------
    signal_list : list of BaseSignal instances
        If sync is set to True, the signals must have the
        same navigation shape, but not necessarily the same signal shape.
    sync : {True, False}, default "True", optional
        If True: the signals will share navigation, all the signals
        must have the same navigation shape for this to work, but not
        necessarily the same signal shape.
    navigator : {"auto", None, "spectrum", "slider", BaseSignal}, optional,
    default "auto"
        See signal.plot docstring for full description.
    navigator_list : {None, list of navigator arguments}, optional, default None
        Set different navigator options for the signals. Must use valid
        navigator arguments: "auto", None, "spectrum", "slider", or a
        HyperSpy Signal. The list must have the same size as signal_list.
        If None, the argument specified in navigator will be used.
    **kwargs
        Any extra keyword arguments are passed to each signal `plot` method.

    Example
    -------

    >>> s_cl = hs.load("coreloss.dm3")
    >>> s_ll = hs.load("lowloss.dm3")
    >>> hs.plot.plot_signals([s_cl, s_ll])

    Specifying the navigator:

    >>> s_cl = hs.load("coreloss.dm3")
    >>> s_ll = hs.load("lowloss.dm3")
    >>> hs.plot.plot_signals([s_cl, s_ll], navigator="slider")

    Specifying the navigator for each signal:

    >>> s_cl = hs.load("coreloss.dm3")
    >>> s_ll = hs.load("lowloss.dm3")
    >>> s_edx = hs.load("edx.dm3")
    >>> s_adf = hs.load("adf.dm3")
    >>> hs.plot.plot_signals(
            [s_cl, s_ll, s_edx], navigator_list=["slider",None,s_adf])

    """

    import hyperspy.signal

    if navigator_list:
        if not (len(signal_list) == len(navigator_list)):
            raise ValueError(
                "signal_list and navigator_list must"
                " have the same size")

    if sync:
        axes_manager_list = []
        for signal in signal_list:
            axes_manager_list.append(signal.axes_manager)

        if not navigator_list:
            navigator_list = []
        if navigator is None:
            navigator_list.extend([None] * len(signal_list))
        elif isinstance(navigator, hyperspy.signal.BaseSignal):
            navigator_list.append(navigator)
            navigator_list.extend([None] * (len(signal_list) - 1))
        elif navigator == "slider":
            navigator_list.append("slider")
            navigator_list.extend([None] * (len(signal_list) - 1))
        elif navigator == "spectrum":
            navigator_list.extend(["spectrum"] * len(signal_list))
        elif navigator == "auto":
            navigator_list.extend(["auto"] * len(signal_list))
        else:
            raise ValueError(
                "navigator must be one of \"spectrum\",\"auto\","
                " \"slider\", None, a Signal instance")

        # Check to see if the spectra have the same navigational shapes
        temp_shape_first = axes_manager_list[0].navigation_shape
        for i, axes_manager in enumerate(axes_manager_list):
            temp_shape = axes_manager.navigation_shape
            if not (temp_shape_first == temp_shape):
                raise ValueError(
                    "The spectra do not have the same navigation shape")
            axes_manager_list[i] = axes_manager.deepcopy()
            if i > 0:
                for axis0, axisn in zip(axes_manager_list[0].navigation_axes,
                                        axes_manager_list[i].navigation_axes):
                    axes_manager_list[i]._axes[axisn.index_in_array] = axis0
            del axes_manager

        for signal, navigator, axes_manager in zip(signal_list,
                                                   navigator_list,
                                                   axes_manager_list):
            signal.plot(axes_manager=axes_manager,
                        navigator=navigator,
                        **kwargs)

    # If sync is False
    else:
        if not navigator_list:
            navigator_list = []
            navigator_list.extend([navigator] * len(signal_list))
        for signal, navigator in zip(signal_list, navigator_list):
            signal.plot(navigator=navigator,
                        **kwargs)


def _make_heatmap_subplot(spectra, **plot_kwargs):
    from hyperspy._signals.signal2d import Signal2D
    im = Signal2D(spectra.data, axes=spectra.axes_manager._get_axes_dicts())
    im.metadata.General.title = spectra.metadata.General.title
    im.plot(**plot_kwargs)
    return im._plot.signal_plot.ax


def set_xaxis_lims(mpl_ax, hs_axis):
    """
    Set the matplotlib axis limits to match that of a HyperSpy axis.

    Parameters
    ----------
    mpl_ax : :class:`matplotlib.axis.Axis`
        The ``matplotlib`` axis to change.
    hs_axis : :class:`~hyperspy.axes.DataAxis`
        The data axis that contains the values which control the scaling.
    """
    x_axis_lower_lim = hs_axis.axis[0]
    x_axis_upper_lim = hs_axis.axis[-1]
    mpl_ax.set_xlim(x_axis_lower_lim, x_axis_upper_lim)


def _make_overlap_plot(spectra, ax, color="blue", line_style='-'):
    if isinstance(color, str):
        color = [color] * len(spectra)
    if isinstance(line_style, str):
        line_style = [line_style] * len(spectra)
    for spectrum_index, (spectrum, color, line_style) in enumerate(
            zip(spectra, color, line_style)):
        x_axis = spectrum.axes_manager.signal_axes[0]
        spectrum = _transpose_if_required(spectrum, 1)
        ax.plot(x_axis.axis, spectrum.data, color=color, ls=line_style)
        set_xaxis_lims(ax, x_axis)
    _set_spectrum_xlabel(spectra if isinstance(spectra, hs.signals.BaseSignal)
                         else spectra[-1], ax)
    ax.set_ylabel('Intensity')
    ax.autoscale(tight=True)


def _make_cascade_subplot(
        spectra, ax, color="blue", line_style='-', padding=1):
    max_value = 0
    for spectrum in spectra:
        spectrum_yrange = (np.nanmax(spectrum.data) -
                           np.nanmin(spectrum.data))
        if spectrum_yrange > max_value:
            max_value = spectrum_yrange
    if isinstance(color, str):
        color = [color] * len(spectra)
    if isinstance(line_style, str):
        line_style = [line_style] * len(spectra)
    for spectrum_index, (spectrum, color, line_style) in enumerate(
            zip(spectra, color, line_style)):
        x_axis = spectrum.axes_manager.signal_axes[0]
        spectrum = _transpose_if_required(spectrum, 1)
        data_to_plot = ((spectrum.data - spectrum.data.min()) /
                        float(max_value) + spectrum_index * padding)
        ax.plot(x_axis.axis, data_to_plot, color=color, ls=line_style)
        set_xaxis_lims(ax, x_axis)
    _set_spectrum_xlabel(spectra if isinstance(spectra, hs.signals.BaseSignal)
                         else spectra[-1], ax)
    ax.set_yticks([])
    ax.autoscale(tight=True)


def _plot_spectrum(spectrum, ax, color="blue", line_style='-'):
    x_axis = spectrum.axes_manager.signal_axes[0]
    ax.plot(x_axis.axis, spectrum.data, color=color, ls=line_style)
    set_xaxis_lims(ax, x_axis)


def _set_spectrum_xlabel(spectrum, ax):
    x_axis = spectrum.axes_manager.signal_axes[0]
    ax.set_xlabel("%s (%s)" % (x_axis.name, x_axis.units))


def _transpose_if_required(signal, expected_dimension):
    # EDS profiles or maps have signal dimension = 0 and navigation dimension
    # 1 or 2. For convenience, transpose the signal if possible
    if (signal.axes_manager.signal_dimension == 0 and
            signal.axes_manager.navigation_dimension == expected_dimension):
        return signal.T
    else:
        return signal


def plot_images(images,
                cmap=None,
                no_nans=False,
                per_row=3,
                label='auto',
                labelwrap=30,
                suptitle=None,
                suptitle_fontsize=18,
                colorbar='multi',
                centre_colormap="auto",
                scalebar=None,
                scalebar_color='white',
                axes_decor='all',
                padding=None,
                tight_layout=False,
                aspect='auto',
                min_asp=0.1,
                namefrac_thresh=0.4,
                fig=None,
                vmin=None,
                vmax=None,
                **kwargs):
    """Plot multiple images as sub-images in one figure.

    Parameters
    ----------
    images : list of Signal2D or BaseSignal
        `images` should be a list of Signals to plot. For `BaseSignal` with
        navigation dimensions 2 and signal dimension 0, the signal will be
        tranposed to form a `Signal2D`.
        Multi-dimensional images will have each plane plotted as a separate
        image.
        If any of the signal shapes is not suitable, a ValueError will be
        raised.
    cmap : {matplotlib colormap, list, ``'mpl_colors'``}, optional
        The colormap used for the images, by default uses the setting
        ``color map signal`` from the plot preferences. A list of colormaps can
        also be provided, and the images will cycle through them. Optionally,
        the value ``'mpl_colors'`` will cause the cmap to loop through the
        default ``matplotlib`` colors (to match with the default output of the
        :py:func:`~.drawing.utils.plot_spectra` method).
        Note: if using more than one colormap, using the ``'single'``
        option for ``colorbar`` is disallowed.
    no_nans : bool, optional
        If True, set nans to zero for plotting.
    per_row : int, optional
        The number of plots in each row.
    label : {None, str, list of str}, optional
        Control the title labeling of the plotted images.
        If None, no titles will be shown.
        If 'auto' (default), function will try to determine suitable titles
        using Signal2D titles, falling back to the 'titles' option if no good
        short titles are detected.
        Works best if all images to be plotted have the same beginning
        to their titles.
        If 'titles', the title from each image's `metadata.General.title`
        will be used.
        If any other single str, images will be labeled in sequence using
        that str as a prefix.
        If a list of str, the list elements will be used to determine the
        labels (repeated, if necessary).
    labelwrap : int, optional
        Integer specifying the number of characters that will be used on
        one line.
        If the function returns an unexpected blank figure, lower this
        value to reduce overlap of the labels between figures.
    suptitle : str, optional
        Title to use at the top of the figure. If called with label='auto',
        this parameter will override the automatically determined title.
    suptitle_fontsize : int, optional
        Font size to use for super title at top of figure.
    colorbar : {'multi', None, 'single'}
        Controls the type of colorbars that are plotted.
        If None, no colorbar is plotted.
        If 'multi' (default), individual colorbars are plotted for each
        (non-RGB) image
        If 'single', all (non-RGB) images are plotted on the same scale,
        and one colorbar is shown for all
    centre_colormap : {"auto", True, False}
        If True, the centre of the color scheme is set to zero. This is
        particularly useful when using diverging color schemes. If "auto"
        (default), diverging color schemes are automatically centred.
    scalebar : {None, 'all', list of ints}, optional
        If None (or False), no scalebars will be added to the images.
        If 'all', scalebars will be added to all images.
        If list of ints, scalebars will be added to each image specified.
    scalebar_color : str, optional
        A valid MPL color string; will be used as the scalebar color.
    axes_decor : {'all', 'ticks', 'off', None}, optional
        Controls how the axes are displayed on each image; default is 'all'.
        If 'all', both ticks and axis labels will be shown.
        If 'ticks', no axis labels will be shown, but ticks/labels will.
        If 'off', all decorations and frame will be disabled.
        If None, no axis decorations will be shown, but ticks/frame will.
    padding : {None, dict}, optional
        This parameter controls the spacing between images.
        If None, default options will be used.
        Otherwise, supply a dictionary with the spacing options as
        keywords and desired values as values.
        Values should be supplied as used in
        :py:func:`matplotlib.pyplot.subplots_adjust`,
        and can be 'left', 'bottom', 'right', 'top', 'wspace' (width) and
        'hspace' (height).
    tight_layout : bool, optional
        If true, hyperspy will attempt to improve image placement in
        figure using matplotlib's tight_layout.
        If false, repositioning images inside the figure will be left as
        an exercise for the user.
    aspect : {str, numeric}, optional
        If 'auto', aspect ratio is auto determined, subject to min_asp.
        If 'square', image will be forced onto square display.
        If 'equal', aspect ratio of 1 will be enforced.
        If float (or int/long), given value will be used.
    min_asp : float, optional
        Minimum aspect ratio to be used when plotting images.
    namefrac_thresh : float, optional
        Threshold to use for auto-labeling. This parameter controls how
        much of the titles must be the same for the auto-shortening of
        labels to activate. Can vary from 0 to 1. Smaller values
        encourage shortening of titles by auto-labeling, while larger
        values will require more overlap in titles before activing the
        auto-label code.
    fig : mpl figure, optional
        If set, the images will be plotted to an existing MPL figure
    vmin, vmax: scalar, str, None
        If str, formatted as 'xth', use this value to calculate the percentage
        of pixels that are left out of the lower and upper bounds.
        For example, for a vmin of '1th', 1% of the lowest will be ignored to
        estimate the minimum value. Similarly, for a vmax value of '1th', 1%
        of the highest value will be ignored in the estimation of the maximum
        value. It must be in the range [0, 100]
        See :py:func:`numpy.percentile` for more explanation.
        If None, use the percentiles value set in the preferences.
        If float of integer, keep this value as bounds.
    **kwargs, optional
        Additional keyword arguments passed to matplotlib.imshow()

    Returns
    -------
    axes_list : list
        A list of subplot axes that hold the images.

    See Also
    --------
    plot_spectra : Plotting of multiple spectra
    plot_signals : Plotting of multiple signals
    plot_histograms : Compare signal histograms

    Notes
    -----
    `interpolation` is a useful parameter to provide as a keyword
    argument to control how the space between pixels is interpolated. A
    value of ``'nearest'`` will cause no interpolation between pixels.

    `tight_layout` is known to be quite brittle, so an option is provided
    to disable it. Turn this option off if output is not as expected,
    or try adjusting `label`, `labelwrap`, or `per_row`.

    """
    def __check_single_colorbar(cbar):
        if cbar == 'single':
            raise ValueError('Cannot use a single colorbar with multiple '
                             'colormaps. Please check for compatible '
                             'arguments.')

    from hyperspy.drawing.widgets import ScaleBar
    from hyperspy.misc import rgb_tools
    from hyperspy.signal import BaseSignal

    # Check that we have a hyperspy signal
    im = [images] if not isinstance(images, (list, tuple)) else images
    for image in im:
        if not isinstance(image, BaseSignal):
            raise ValueError("`images` must be a list of image signals or a "
                             "multi-dimensional signal."
                             " " + repr(type(images)) + " was given.")

    # For list of EDS maps, transpose the BaseSignal
    if isinstance(images, (list, tuple)):
        images = [_transpose_if_required(image, 2) for image in images]

    # If input is >= 1D signal (e.g. for multi-dimensional plotting),
    # copy it and put it in a list so labeling works out as (x,y) when plotting
    if isinstance(images,
                  BaseSignal) and images.axes_manager.navigation_dimension > 0:
        images = [images._deepcopy_with_new_data(images.data)]

    n = 0
    for i, sig in enumerate(images):
        if sig.axes_manager.signal_dimension != 2:
            raise ValueError("This method only plots signals that are images. "
                             "The signal dimension must be equal to 2. "
                             "The signal at position " + repr(i) +
                             " was " + repr(sig) + ".")
        # increment n by the navigation size, or by 1 if the navigation size is
        # <= 0
        n += (sig.axes_manager.navigation_size
              if sig.axes_manager.navigation_size > 0
              else 1)

    # If no cmap given, get default colormap from pyplot:
    if cmap is None:
        cmap = [preferences.Plot.cmap_signal]
    elif cmap == 'mpl_colors':
        for n_color, c in enumerate(mpl.rcParams['axes.prop_cycle']):
            make_cmap(colors=['#000000', c['color']],
                      name='mpl{}'.format(n_color))
        cmap = ['mpl{}'.format(i) for i in
                range(len(mpl.rcParams['axes.prop_cycle']))]
        __check_single_colorbar(colorbar)
    # cmap is list, tuple, or something else iterable (but not string):
    elif hasattr(cmap, '__iter__') and not isinstance(cmap, str):
        try:
            cmap = [c.name for c in cmap]  # convert colormap to string
        except AttributeError:
            cmap = [c for c in cmap]   # c should be string if not colormap
        __check_single_colorbar(colorbar)
    elif isinstance(cmap, mpl.colors.Colormap):
        cmap = [cmap.name]   # convert single colormap to list with string
    elif isinstance(cmap, str):
        cmap = [cmap]  # cmap is single string, so make it a list
    else:
        # Didn't understand cmap input, so raise error
        raise ValueError('The provided cmap value was not understood. Please '
                         'check input values.')

    # If any of the cmaps given are diverging, and auto-centering, set the
    # appropriate flag:
    if centre_colormap == "auto":
        centre_colormaps = []
        for c in cmap:
            if c in MPL_DIVERGING_COLORMAPS:
                centre_colormaps.append(True)
            else:
                centre_colormaps.append(False)
    # if it was True, just convert to list
    elif centre_colormap:
        centre_colormaps = [True]
    # likewise for false
    elif not centre_colormap:
        centre_colormaps = [False]

    # finally, convert lists to cycle generators for adaptive length:
    centre_colormaps = itertools.cycle(centre_colormaps)
    cmap = itertools.cycle(cmap)

    # Sort out the labeling:
    div_num = 0
    all_match = False
    shared_titles = False
    user_labels = False

    if label is None:
        pass
    elif label == 'auto':
        # Use some heuristics to try to get base string of similar titles

        label_list = [x.metadata.General.title for x in images]

        # Find the shortest common string between the image titles
        # and pull that out as the base title for the sequence of images
        # array in which to store arrays
        res = np.zeros((len(label_list), len(label_list[0]) + 1))
        res[:, 0] = 1

        # j iterates the strings
        for j in range(len(label_list)):
            # i iterates length of substring test
            for i in range(1, len(label_list[0]) + 1):
                # stores whether or not characters in title match
                res[j, i] = label_list[0][:i] in label_list[j]

        # sum up the results (1 is True, 0 is False) and create
        # a substring based on the minimum value (this will be
        # the "smallest common string" between all the titles
        if res.all():
            basename = label_list[0]
            div_num = len(label_list[0])
            all_match = True
        else:
            div_num = int(min(np.sum(res, 1)))
            basename = label_list[0][:div_num - 1]
            all_match = False

        # trim off any '(' or ' ' characters at end of basename
        if div_num > 1:
            while True:
                if basename[len(basename) - 1] == '(':
                    basename = basename[:-1]
                elif basename[len(basename) - 1] == ' ':
                    basename = basename[:-1]
                else:
                    break

        # namefrac is ratio of length of basename to the image name
        # if it is high (e.g. over 0.5), we can assume that all images
        # share the same base
        if len(label_list[0]) > 0:
            namefrac = float(len(basename)) / len(label_list[0])
        else:
            # If label_list[0] is empty, it means there was probably no
            # title set originally, so nothing to share
            namefrac = 0

        if namefrac > namefrac_thresh:
            # there was a significant overlap of label beginnings
            shared_titles = True
            # only use new suptitle if one isn't specified already
            if suptitle is None:
                suptitle = basename

        else:
            # there was not much overlap, so default back to 'titles' mode
            shared_titles = False
            label = 'titles'
            div_num = 0

    elif label == 'titles':
        # Set label_list to each image's pre-defined title
        label_list = [x.metadata.General.title for x in images]

    elif isinstance(label, str):
        # Set label_list to an indexed list, based off of label
        label_list = [label + " " + repr(num) for num in range(n)]

    elif isinstance(label, list) and all(
            isinstance(x, str) for x in label):
        label_list = label
        user_labels = True
        # If list of labels is longer than the number of images, just use the
        # first n elements
        if len(label_list) > n:
            del label_list[n:]
        if len(label_list) < n:
            label_list *= (n // len(label_list)) + 1
            del label_list[n:]

    else:
        raise ValueError("Did not understand input of labels.")

    # Determine appropriate number of images per row
    rows = int(np.ceil(n / float(per_row)))
    if n < per_row:
        per_row = n

    # Set overall figure size and define figure (if not pre-existing)
    if fig is None:
        k = max(plt.rcParams['figure.figsize']) / max(per_row, rows)
        f = plt.figure(figsize=(tuple(k * i for i in (per_row, rows))))
    else:
        f = fig

    # Initialize list to hold subplot axes
    axes_list = []

    # Initialize list of rgb tags
    isrgb = [False] * len(images)

    # Check to see if there are any rgb images in list
    # and tag them using the isrgb list
    for i, img in enumerate(images):
        if rgb_tools.is_rgbx(img.data):
            isrgb[i] = True

    # Determine how many non-rgb images there are
    non_rgb = list(itertools.compress(images, [not j for j in isrgb]))
    if len(non_rgb) == 0 and colorbar is not None:
        colorbar = None
        warnings.warn("Sorry, colorbar is not implemented for RGB images.")

    # Check if we need to add a scalebar for some of the images
    if isinstance(scalebar, list) and all(isinstance(x, int)
                                          for x in scalebar):
        scalelist = True
    else:
        scalelist = False

    def check_list_length(arg, arg_name):
        if isinstance(arg, (list, tuple)):
            if len(arg) != n:
                _logger.warning(f'The provided {arg_name} values are ignored '
                                'because the length of the list does not '
                                'match the number of images')
                arg = [None] * n
        return arg

    # Find global min and max values of all the non-rgb images for use with
    # 'single' scalebar, otherwise define this value later.
    if colorbar == 'single':
        # check that vmin and vmax are not list
        if any([isinstance(v, (tuple, list)) for v in [vmin, vmax]]):
            _logger.warning('The provided vmin or vmax value are ignored '
                            'because it needs to be a scalar or a str '
                            'to be compatible with a single colorbar. '
                            'The default values are used instead.')
            vmin, vmax = None, None
        vmin_max = np.array(
            [contrast_stretching(i.data, vmin, vmax) for i in non_rgb])
        _vmin, _vmax = vmin_max[:, 0].min(), vmin_max[:, 1].max()
        if next(centre_colormaps):
            _vmin, _vmax = centre_colormap_values(_vmin, _vmax)

    else:
        vmin = check_list_length(vmin, "vmin")
        vmax = check_list_length(vmax, "vmax")

    idx = 0
    ax_im_list = [0] * len(isrgb)

    # Replot: create a list to store references to the images
    replot_ims = []

    # Loop through each image, adding subplot for each one
    for i, ims in enumerate(images):
        # Get handles for the signal axes and axes_manager
        axes_manager = ims.axes_manager
        if axes_manager.navigation_dimension > 0:
            ims = ims._deepcopy_with_new_data(ims.data)
        for j, im in enumerate(ims):
            ax = f.add_subplot(rows, per_row, idx + 1)
            axes_list.append(ax)
            data = im.data
            centre = next(centre_colormaps)   # get next value for centreing

            # Enable RGB plotting
            if rgb_tools.is_rgbx(data):
                data = rgb_tools.rgbx2regular_array(data, plot_friendly=True)
                _vmin, _vmax = None, None
            elif colorbar != 'single':
                _vmin = vmin[idx] if isinstance(vmin, (tuple, list)) else vmin
                _vmax = vmax[idx] if isinstance(vmax, (tuple, list)) else vmax
                _vmin, _vmax = contrast_stretching(data, _vmin, _vmax)
                if centre:
                    _vmin, _vmax = centre_colormap_values(_vmin, _vmax)

            # Remove NaNs (if requested)
            if no_nans:
                data = np.nan_to_num(data)

            # Get handles for the signal axes and axes_manager
            axes_manager = im.axes_manager
            axes = axes_manager.signal_axes

            # Set dimensions of images
            xaxis = axes[0]
            yaxis = axes[1]

            extent = (
                xaxis.low_value,
                xaxis.high_value,
                yaxis.high_value,
                yaxis.low_value,
            )

            if not isinstance(aspect, (int, float)) and aspect not in [
                    'auto', 'square', 'equal']:
                _logger.warning("Did not understand aspect ratio input. "
                                "Using 'auto' as default.")
                aspect = 'auto'

            if aspect == 'auto':
                if float(yaxis.size) / xaxis.size < min_asp:
                    factor = min_asp * float(xaxis.size) / yaxis.size
                elif float(yaxis.size) / xaxis.size > min_asp ** -1:
                    factor = min_asp ** -1 * float(xaxis.size) / yaxis.size
                else:
                    factor = 1
                asp = np.abs(factor * float(xaxis.scale) / yaxis.scale)
            elif aspect == 'square':
                asp = abs(extent[1] - extent[0]) / abs(extent[3] - extent[2])
            elif aspect == 'equal':
                asp = 1
            elif isinstance(aspect, (int, float)):
                asp = aspect
            if 'interpolation' not in kwargs.keys():
                kwargs['interpolation'] = 'nearest'

            # Plot image data, using _vmin and _vmax to set bounds,
            # or allowing them to be set automatically if using individual
            # colorbars
            kwargs.update({'cmap':next(cmap), 'extent':extent, 'aspect':asp})
            axes_im = ax.imshow(data, vmin=_vmin, vmax=_vmax, **kwargs)
            ax_im_list[i] = axes_im

            # If an axis trait is undefined, shut off :
            if (xaxis.units == t.Undefined or yaxis.units == t.Undefined or
                xaxis.name == t.Undefined or yaxis.name == t.Undefined):
                if axes_decor == 'all':
                    _logger.warning(
                        'Axes labels were requested, but one '
                        'or both of the '
                        'axes units and/or name are undefined. '
                        'Axes decorations have been set to '
                        '\'ticks\' instead.')
                    axes_decor = 'ticks'
            # If all traits are defined, set labels as appropriate:
            else:
                ax.set_xlabel(axes[0].name + " axis (" + axes[0].units + ")")
                ax.set_ylabel(axes[1].name + " axis (" + axes[1].units + ")")

            if label:
                if all_match:
                    title = ''
                elif shared_titles:
                    title = label_list[i][div_num - 1:]
                else:
                    if len(ims) == n:
                        # This is true if we are plotting just 1
                        # multi-dimensional Signal2D
                        title = label_list[idx]
                    elif user_labels:
                        title = label_list[idx]
                    else:
                        title = label_list[i]

                if ims.axes_manager.navigation_size > 1 and not user_labels:
                    title += " %s" % str(ims.axes_manager.indices)

                ax.set_title(textwrap.fill(title, labelwrap))

            # Set axes decorations based on user input
            set_axes_decor(ax, axes_decor)

            # If using independent colorbars, add them
            if colorbar == 'multi' and not isrgb[i]:
                div = make_axes_locatable(ax)
                cax = div.append_axes("right", size="5%", pad=0.05)
                plt.colorbar(axes_im, cax=cax)

            # Add scalebars as necessary
            if (scalelist and idx in scalebar) or scalebar == 'all':
                ax.scalebar = ScaleBar(
                    ax=ax,
                    units=axes[0].units,
                    color=scalebar_color,
                )
            # Replot: store references to the images
            replot_ims.append(im)

            idx += 1

    # If using a single colorbar, add it, and do tight_layout, ensuring that
    # a colorbar is only added based off of non-rgb Images:
    if colorbar == 'single':
        foundim = None
        for i in range(len(isrgb)):
            if (not isrgb[i]) and foundim is None:
                foundim = i

        if foundim is not None:
            f.subplots_adjust(right=0.8)
            cbar_ax = f.add_axes([0.9, 0.1, 0.03, 0.8])
            f.colorbar(ax_im_list[foundim], cax=cbar_ax)
            if tight_layout:
                # tight_layout, leaving room for the colorbar
                plt.tight_layout(rect=[0, 0, 0.9, 1])
        elif tight_layout:
            plt.tight_layout()

    elif tight_layout:
        plt.tight_layout()

    # Set top bounds for shared titles and add suptitle
    if suptitle:
        f.subplots_adjust(top=0.85)
        f.suptitle(suptitle, fontsize=suptitle_fontsize)

    # If we want to plot scalebars, loop through the list of axes and add them
    if scalebar is None or scalebar is False:
        # Do nothing if no scalebars are called for
        pass
    elif scalebar == 'all':
        # scalebars were taken care of in the plotting loop
        pass
    elif scalelist:
        # scalebars were taken care of in the plotting loop
        pass
    else:
        raise ValueError("Did not understand scalebar input. Must be None, "
                         "'all', or list of ints.")

    # Adjust subplot spacing according to user's specification
    if padding is not None:
        plt.subplots_adjust(**padding)

    # Replot: connect function
    def on_dblclick(event):
        # On the event of a double click, replot the selected subplot
        if not event.inaxes:
            return
        if not event.dblclick:
            return
        subplots = [axi for axi in f.axes if isinstance(axi, mpl.axes.Subplot)]
        inx = list(subplots).index(event.inaxes)
        im = replot_ims[inx]

        # Use some of the info in the subplot
        cm = subplots[inx].images[0].get_cmap()
        clim = subplots[inx].images[0].get_clim()

        sbar = False
        if (scalelist and inx in scalebar) or scalebar == 'all':
            sbar = True

        im.plot(colorbar=bool(colorbar),
                vmin=clim[0],
                vmax=clim[1],
                no_nans=no_nans,
                aspect=asp,
                scalebar=sbar,
                scalebar_color=scalebar_color,
                cmap=cm)

    f.canvas.mpl_connect('button_press_event', on_dblclick)

    return axes_list


def set_axes_decor(ax, axes_decor):
    if axes_decor == 'off':
        ax.axis('off')
    elif axes_decor == 'ticks':
        ax.set_xlabel('')
        ax.set_ylabel('')
    elif axes_decor == 'all':
        pass
    elif axes_decor is None:
        ax.set_xlabel('')
        ax.set_ylabel('')
        ax.set_xticklabels([])
        ax.set_yticklabels([])


def make_cmap(colors, name='my_colormap', position=None,
              bit=False, register=True):
    """
    Create a matplotlib colormap with customized colors, optionally registering
    it with matplotlib for simplified use.

    Adapted from Chris Slocum's code at:
    https://github.com/CSlocumWX/custom_colormap/blob/master/custom_colormaps.py
    and used under the terms of that code's BSD-3 license

    Parameters
    ----------
    colors : iterable
        list of either tuples containing rgb values, or html strings
        Colors should be arranged so that the first color is the lowest
        value for the colorbar and the last is the highest.
    name : str
        name of colormap to use when registering with matplotlib
    position : {None, iterable}, optional
        list containing the values (from [0,1]) that dictate the position
        of each color within the colormap. If None (default), the colors
        will be equally-spaced within the colorbar.
    bit : bool, optional
        True if RGB colors are given in 8-bit [0 to 255] or False if given
        in arithmetic basis [0 to 1] (default).
    register : bool, optional
        Wwitch to control whether or not to register the custom colormap
        with matplotlib in order to enable use by just the name string.
    """
    bit_rgb = np.linspace(0, 1, 256)

    if position is None:
        position = np.linspace(0, 1, len(colors))
    else:
        if len(position) != len(colors):
            raise ValueError("Position length must be the same as colors")
        elif position[0] != 0 or position[-1] != 1:
            raise ValueError("Position must start with 0 and end with 1")

    cdict = {'red': [], 'green': [], 'blue': []}

    for pos, color in zip(position, colors):
        if isinstance(color, str):
            color = mpl.colors.to_rgb(color)
        elif bit:
            color = (bit_rgb[color[0]],
                     bit_rgb[color[1]],
                     bit_rgb[color[2]])

        cdict['red'].append((pos, color[0], color[0]))
        cdict['green'].append((pos, color[1], color[1]))
        cdict['blue'].append((pos, color[2], color[2]))

    cmap = mpl.colors.LinearSegmentedColormap(name, cdict, 256)

    if register:
        mpl.cm.register_cmap(name, cmap)
    return cmap


def plot_spectra(
        spectra,
        style='overlap',
        color=None,
        line_style=None,
        padding=1.,
        legend=None,
        legend_picking=True,
        legend_loc='upper right',
        fig=None,
        ax=None,
        auto_update=None,
        **kwargs):
    """Plot several spectra in the same figure.

    Parameters
    ----------
    spectra : list of Signal1D or BaseSignal
        Ordered spectra list of signal to plot. If `style` is "cascade" or
        "mosaic", the spectra can have different size and axes. For `BaseSignal`
        with navigation dimensions 1 and signal dimension 0, the signal will be
        tranposed to form a `Signal1D`.
    style : {'overlap', 'cascade', 'mosaic', 'heatmap'}
        The style of the plot.
    color : {None, matplotlib color, list of colors}, optional
        Sets the color of the lines of the plots (no action on 'heatmap').
        For a list, if its length is less than the number of spectra to plot,
        the colors will be cycled. If `None`, use default matplotlib color
        cycle.
    line_style: {None, matplotlib line style, list of line_styles}, optional
        Sets the line style of the plots (no action on 'heatmap').
        The main line style are '-','--','steps','-.',':'.
        For a list, if its length is less than the number of
        spectra to plot, line_style will be cycled. If
        If `None`, use continuous lines, eg: ('-','--','steps','-.',':').
    padding : float, optional, default 1.0
        Option for "cascade". 1 guarantees that there is no overlapping.
        However, in many cases, a value between 0 and 1 can produce a tighter
        plot without overlapping. Negative values have the same effect but
        reverse the order of the spectra without reversing the order of the
        colors.
    legend: {None, list of str, 'auto'}, optional
       If list of string, legend for "cascade" or title for "mosaic" is
       displayed. If 'auto', the title of each spectra (metadata.General.title)
       is used.
    legend_picking: bool, optional
        If True (default), a spectrum can be toggled on and off by clicking on
        the legended line.
    legend_loc : {str, int}, optional
        This parameter controls where the legend is placed on the figure;
        see the pyplot.legend docstring for valid values
    fig : {None, matplotlib figure}, optional
        If None, a default figure will be created. Specifying fig will
        not work for the 'heatmap' style.
    ax : {none, matplotlib ax (subplot)}, optional
        If None, a default ax will be created. Will not work for 'mosaic'
        or 'heatmap' style.
    auto_update : bool or None
        If True, the plot will update when the data are changed. Only supported
        with style='overlap' and a list of signal with navigation dimension 0.
        If None (default), update the plot only for style='overlap'.
    **kwargs, optional
        Keywords arguments passed to :py:func:`matplotlib.pyplot.figure` or
        :py:func:`matplotlib.pyplot.subplots` if style='mosaic'.
        Has no effect on 'heatmap' style.

    Example
    -------
    >>> s = hs.load("some_spectra")
    >>> hs.plot.plot_spectra(s, style='cascade', color='red', padding=0.5)

    To save the plot as a png-file

    >>> hs.plot.plot_spectra(s).figure.savefig("test.png")

    Returns
    -------
    ax: matplotlib axes or list of matplotlib axes
        An array is returned when `style` is "mosaic".

    """
    import hyperspy.signal

    def _reverse_legend(ax_, legend_loc_):
        """
        Reverse the ordering of a matplotlib legend (to be more consistent
        with the default ordering of plots in the 'cascade' and 'overlap'
        styles.

        Parameters
        ----------
        ax_: matplotlib axes

        legend_loc_: {str, int}
            This parameter controls where the legend is placed on the
            figure; see the pyplot.legend docstring for valid values.
        """
        l = ax_.get_legend()
        labels = [lb.get_text() for lb in list(l.get_texts())]
        handles = l.legendHandles
        ax_.legend(handles[::-1], labels[::-1], loc=legend_loc_)

    # Before v1.3 default would read the value from prefereces.
    if style == "default":
        style = "overlap"

    if color is not None:
        if isinstance(color, str):
            color = itertools.cycle([color])
        elif hasattr(color, "__iter__"):
            color = itertools.cycle(color)
        else:
            raise ValueError("Color must be None, a valid matplotlib color "
                             "string, or a list of valid matplotlib colors.")
    else:
        color = itertools.cycle(
                plt.rcParams['axes.prop_cycle'].by_key()["color"])

    if line_style is not None:
        if isinstance(line_style, str):
            line_style = itertools.cycle([line_style])
        elif hasattr(line_style, "__iter__"):
            line_style = itertools.cycle(line_style)
        else:
            raise ValueError("line_style must be None, a valid matplotlib "
                             "line_style string or a list of valid matplotlib "
                             "line_style.")
    else:
        line_style = ['-'] * len(spectra)

    if legend is not None:
        if isinstance(legend, str):
            if legend == 'auto':
                legend = [spec.metadata.General.title for spec in spectra]
            else:
                raise ValueError("legend must be None, 'auto' or a list of "
                                 "strings.")

    if style == 'overlap':
        if fig is None:
            fig = plt.figure(**kwargs)
        if ax is None:
            ax = fig.add_subplot(111)
        _make_overlap_plot(spectra,
                           ax,
                           color=color,
                           line_style=line_style,)
        if legend is not None:
            ax.legend(legend, loc=legend_loc)
            _reverse_legend(ax, legend_loc)
            if legend_picking is True:
                animate_legend(fig=fig, ax=ax)
    elif style == 'cascade':
        if fig is None:
            fig = plt.figure(**kwargs)
        if ax is None:
            ax = fig.add_subplot(111)
        _make_cascade_subplot(spectra,
                              ax,
                              color=color,
                              line_style=line_style,
                              padding=padding)
        if legend is not None:
            plt.legend(legend, loc=legend_loc)
            _reverse_legend(ax, legend_loc)
    elif style == 'mosaic':
        default_fsize = plt.rcParams["figure.figsize"]
        figsize = (default_fsize[0], default_fsize[1] * len(spectra))
        fig, subplots = plt.subplots(
            len(spectra), 1, figsize=figsize, **kwargs)
        if legend is None:
            legend = [legend] * len(spectra)
        for spectrum, ax, color, line_style, legend in zip(
                spectra, subplots, color, line_style, legend):
            spectrum = _transpose_if_required(spectrum, 1)
            _plot_spectrum(spectrum, ax, color=color, line_style=line_style)
            ax.set_ylabel('Intensity')
            if legend is not None:
                ax.set_title(legend)
            if not isinstance(spectra, hyperspy.signal.BaseSignal):
                _set_spectrum_xlabel(spectrum, ax)
        if isinstance(spectra, hyperspy.signal.BaseSignal):
            _set_spectrum_xlabel(spectrum, ax)
        fig.tight_layout()

    elif style == 'heatmap':
        if not isinstance(spectra, hyperspy.signal.BaseSignal):
            import hyperspy.utils
            spectra = [_transpose_if_required(spectrum, 1) for spectrum in
                       spectra]
            spectra = hyperspy.utils.stack(spectra)
        with spectra.unfolded():
            ax = _make_heatmap_subplot(spectra)
            ax.set_ylabel('Spectra')
    ax = ax if style != "mosaic" else subplots

    def update_line(spectrum, line):
        x_axis = spectrum.axes_manager[-1].axis
        line.set_data(x_axis, spectrum.data)
        fig = line.get_figure()
        ax = fig.get_axes()[0]
        # `relim` needs to be called before `autoscale_view`
        ax.relim()
        ax.autoscale_view()
        fig.canvas.draw()

    if auto_update is None and style == 'overlap':
        auto_update = True

    if auto_update:
        if style != 'overlap':
            raise ValueError("auto_update=True is only supported with "
                             "style='overlap'.")

        for spectrum, line in zip(spectra, ax.get_lines()):
            f = partial(update_line, spectrum, line)
            spectrum.events.data_changed.connect(f, [])
            # disconnect event when closing figure
            disconnect = partial(spectrum.events.data_changed.disconnect, f)
            on_figure_window_close(fig, disconnect)

    return ax


def animate_legend(fig=None, ax=None):
    """Animate the legend of a figure.

    A spectrum can be toggled on and off by clicking on the line in the legend.

    Parameters
    ----------

    fig: {None, matplotlib.figure}, optional
        If None pick the current figure using "plt.gcf".
    ax:  {None, matplotlib.axes}, optional
        If None pick the current axes using "plt.gca".

    Note
    ----

    Code inspired from legend_picking.py in the matplotlib gallery.

    """
    if fig is None:
        fig = plt.gcf()
    if ax is None:
        ax = plt.gca()
    lines = ax.lines[::-1]
    lined = dict()
    leg = ax.get_legend()
    for legline, origline in zip(leg.get_lines(), lines):
        legline.set_pickradius(5)  # 5 pts tolerance
        legline.set_picker(True)
        lined[legline] = origline

    def onpick(event):
        # on the pick event, find the orig line corresponding to the
        # legend proxy line, and toggle the visibility
        legline = event.artist
        if legline.axes == ax:
            origline = lined[legline]
            vis = not origline.get_visible()
            origline.set_visible(vis)
            # Change the alpha on the line in the legend so we can see what lines
            # have been toggled
            if vis:
                legline.set_alpha(1.0)
            else:
                legline.set_alpha(0.2)
            fig.canvas.draw_idle()

    fig.canvas.mpl_connect('pick_event', onpick)


def plot_histograms(signal_list,
                    bins='fd',
                    range_bins=None,
                    color=None,
                    line_style=None,
                    legend='auto',
                    fig=None,
                    **kwargs):
    """Plot the histogram of every signal in the list in one figure.

    This function creates a histogram for each signal and plots the list with
    the `utils.plot.plot_spectra` function.

    Parameters
    ----------
    signal_list : iterable
        Ordered list of spectra to plot. If `style` is "cascade" or "mosaic",
        the spectra can have different size and axes.
    bins : {int, list, str}, optional
        If bins is a string, then it must be one of:
        'knuth' : use Knuth's rule to determine bins,
        'scott' : use Scott's rule to determine bins,
        'fd' : use the Freedman-diaconis rule to determine bins,
        'blocks' : use bayesian blocks for dynamic bin widths.
    range_bins : {None, tuple}, optional
        The minimum and maximum range for the histogram. If not specified,
        it will be (x.min(), x.max()).
    color : {None, valid matplotlib color, list of colors}, optional
        Sets the color of the lines of the plots. For a list, if its length is
        less than the number of spectra to plot, the colors will be cycled.
        If `None`, use default matplotlib color cycle.
    line_style: {None, valid matplotlib line style, list of line styles},
    optional
        The main line styles are '-','--','steps','-.',':'.
        For a list, if its length is less than the number of
        spectra to plot, line_style will be cycled.
        If `None`, use continuous lines, eg: ('-','--','steps','-.',':')
    legend: {None, list of str, 'auto'}, optional
       Display a legend. If 'auto', the title of each spectra
       (metadata.General.title) is used.
    legend_picking: bool, optional
        If true, a spectrum can be toggled on and off by clicking on
        the line in the legend.
    fig : {None, matplotlib figure}, optional
        If None, a default figure will be created.
    **kwargs
        other keyword arguments (weight and density) are described in
        :py:func:`numpy.histogram`.

    Example
    -------
    Histograms of two random chi-square distributions.

    >>> img = hs.signals.Signal2D(np.random.chisquare(1,[10,10,100]))
    >>> img2 = hs.signals.Signal2D(np.random.chisquare(2,[10,10,100]))
    >>> hs.plot.plot_histograms([img,img2],legend=['hist1','hist2'])

    Returns
    -------
    ax: matplotlib axes or list of matplotlib axes
        An array is returned when `style` is "mosaic".

    """
    hists = []
    for obj in signal_list:
        hists.append(obj.get_histogram(bins=bins,
                                       range_bins=range_bins, **kwargs))
    if line_style is None:
        line_style = 'steps'
    return plot_spectra(hists, style='overlap', color=color,
                        line_style=line_style, legend=legend, fig=fig)


def picker_kwargs(value, kwargs={}):
    # picker is deprecated in favor of pickradius
    if LooseVersion(mpl.__version__) >= LooseVersion("3.3.0"):
        kwargs.update({'pickradius': value, 'picker':True})
    else:
        kwargs['picker'] = value

    return kwargs




# -*- coding: utf-8 -*-
# Copyright 2007-2020 The HyperSpy developers
#
# This file is part of  HyperSpy.
#
#  HyperSpy is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
#  HyperSpy is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with  HyperSpy.  If not, see <http://www.gnu.org/licenses/>.


import logging

from hyperspy.io_plugins import (
    blockfile,
    bruker,
    dens,
    digital_micrograph,
    edax,
    emd,
    empad,
    fei,
    hspy,
    image,
    mrc,
    msa,
    nexus,
    phenom,
    protochips,
    ripple,
    semper_unf,
    sur,
    tiff,
)

io_plugins = [
    blockfile,
    bruker,
    dens,
    digital_micrograph,
    edax,
    emd,
    empad,
    fei,
    hspy,
    image,
    mrc,
    msa,
    nexus,
    phenom,
    protochips,
    ripple,
    semper_unf,
    sur,
    tiff,
]


_logger = logging.getLogger(__name__)


try:
    from hyperspy.io_plugins import netcdf

    io_plugins.append(netcdf)
except ImportError:
    # NetCDF is obsolete and is only provided for users who have
    # old EELSLab files. Therefore, we silently ignore if missing.
    pass

try:
    from hyperspy.io_plugins import usid_hdf5

    io_plugins.append(usid_hdf5)
except ImportError:
    _logger.info(
        "The USID IO plugin is not available because "
        "the pyUSID or sidpy packages are not installed."
    )

try:
    from hyperspy.io_plugins import mrcz

    io_plugins.append(mrcz)
except ImportError:
    _logger.info(
        "The mrcz IO plugin is not available because "
        "the mrcz package is not installed."
    )


default_write_ext = set()
for plugin in io_plugins:
    if plugin.writes:
        default_write_ext.add(plugin.file_extensions[plugin.default_extension])



# -*- coding: utf-8 -*-
# Copyright 2007-2020 The HyperSpy developers
#
# This file is part of HyperSpy.
#
# HyperSpy is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# HyperSpy is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with HyperSpy.  If not, see <http://www.gnu.org/licenses/>.

import os

import numpy as np
import pytest

import hyperspy.api as hs
from hyperspy.io_plugins.protochips import ProtochipsCSV, invalid_file_error

testdirpath = os.path.dirname(__file__)
dirpath = os.path.join(testdirpath, 'protochips_data')

# To generate a new reference numpy file
generate_numpy_file = False


def create_numpy_file(filename, obj):
    gen = (obj._data_dictionary[key] for key in obj.logged_quantity_name_list)
    data = np.vstack(list(gen))
    np.save(filename, data.T)

#######################
# Protochips gas cell #
#######################


def test_read_protochips_gas_cell():
    filename = os.path.join(dirpath, 'protochips_gas_cell.csv')
    s = hs.load(filename)
    assert len(s) == 5
    assert (s[0].metadata.General.title ==
            'Holder Temperature (Degrees C)')
    assert s[0].metadata.Signal.signal_type == ''
    assert s[0].metadata.Signal.quantity == 'Temperature (Degrees C)'
    assert s[1].metadata.General.title == 'Holder Pressure (Torr)'
    assert s[1].metadata.Signal.signal_type == ''
    assert s[1].metadata.Signal.quantity == 'Pressure (Torr)'
    assert s[2].metadata.General.title == 'Tank1 Pressure (Torr)'
    assert s[2].metadata.Signal.signal_type == ''
    assert s[2].metadata.Signal.quantity == 'Pressure (Torr)'
    assert s[3].metadata.General.title == 'Tank2 Pressure (Torr)'
    assert s[3].metadata.Signal.signal_type == ''
    assert s[3].metadata.Signal.quantity == 'Pressure (Torr)'
    assert s[4].metadata.General.title == 'Vacuum Tank Pressure (Torr)'
    assert s[4].metadata.Signal.signal_type == ''
    assert s[4].metadata.Signal.quantity == 'Pressure (Torr)'


def get_datetime(dt):
    dt_np = np.datetime64(dt)
    dt_str = np.datetime_as_string(dt_np)
    date, time = dt_str.split('T')
    return date, time, dt_np


datetime_gas_cell = get_datetime('2014-12-15T19:07:04.165000')
datetime_gas_cell_no_user = get_datetime('2016-10-17T16:59:20.391000')


def test_loading_random_csv_file():
    filename = os.path.join(dirpath, 'random_csv_file.csv')
    with pytest.raises(IOError) as cm:
        ProtochipsCSV(filename)
        cm.match(invalid_file_error)


def test_loading_invalid_protochips_file():
    filename = os.path.join(dirpath, 'invalid_protochips_file.csv')
    with pytest.raises(IOError) as cm:
        hs.load(filename)
        cm.match(invalid_file_error)


class TestProtochipsGasCellCSV():

    def setup_method(self, method):
        filename = os.path.join(dirpath, 'protochips_gas_cell.csv')
        self.s_list = hs.load(filename)

    def test_read_metadata(self):
        date, time, dt_np = datetime_gas_cell
        for s in self.s_list:
            assert s.metadata.General.date == date
            assert s.metadata.General.time == time
            assert s.axes_manager[0].units == 's'
            np.testing.assert_allclose(s.axes_manager[0].scale, 0.25995, atol=1E-5)
            assert s.axes_manager[0].offset == 0

    def test_read_original_metadata(self):
        om = self.s_list[0].original_metadata.Protochips_header
        assert (om.Calibration_file_path == "The calibration files names"
                " are saved in the 'Original notes' array of the "
                "original metadata.")
        assert om.Holder_Pressure_units == 'Torr'
        assert om.Holder_Temperature_units == 'Degrees C'
        assert om.Start_time == datetime_gas_cell[2]
        assert om.Holder_Pressure_units == 'Torr'
        assert om.Tank1_Pressure_units == 'Torr'
        assert om.Tank2_Pressure_units == 'Torr'
        assert om.Vacuum_Tank_Pressure_units == 'Torr'
        assert om.Time_units == 'Milliseconds'
        assert om.User == 'eric'


class TestProtochipsGasCellCSVNoUser():

    def setup_method(self, method):
        filename = os.path.join(dirpath, 'protochips_gas_cell_no_user.csv')
        self.s_list = hs.load(filename)

    def test_read_metadata(self):
        date, time, dt_np = datetime_gas_cell_no_user
        for s in self.s_list:
            assert s.metadata.General.date == date
            assert s.metadata.General.time == time
            assert s.axes_manager[0].units == 's'
            np.testing.assert_allclose(s.axes_manager[0].scale, 0.26029, atol=1E-5)
            assert s.axes_manager[0].offset == 0

    def test_read_original_metadata(self):
        om = self.s_list[0].original_metadata.Protochips_header
        assert (om.Calibration_file_path == "The calibration files names"
                " are saved in the 'Original notes' array of the "
                "original metadata.")
        assert om.Holder_Pressure_units == 'Torr'
        assert om.Holder_Temperature_units == 'Degrees C'
        assert om.Start_time == datetime_gas_cell_no_user[2]
        assert om.Holder_Pressure_units == 'Torr'
        assert om.Tank1_Pressure_units == 'Torr'
        assert om.Tank2_Pressure_units == 'Torr'
        assert om.Vacuum_Tank_Pressure_units == 'Torr'
        assert om.Time_units == 'Milliseconds'


class TestProtochipsGasCellCSVReader():

    def setup_method(self, method):
        self.filename = os.path.join(dirpath, 'protochips_gas_cell.csv')
        self.pgc = ProtochipsCSV(self.filename)
        if generate_numpy_file:
            create_numpy_file(self.filename.replace('.csv', '.npy'), self.pgc)

    def test_read_column_name(self):
        assert self.pgc.column_name == ['Time', 'Notes',
                                        'Holder Temperature',
                                        'Holder Pressure',
                                        'Tank1 Pressure',
                                        'Tank2 Pressure',
                                        'Vacuum Tank Pressure']

    def test_read_start_datetime(self):
        assert self.pgc.start_datetime == datetime_gas_cell[2]

    def test_read_data(self):
        gen = (self.pgc._data_dictionary[key]
               for key in self.pgc.logged_quantity_name_list)
        data = np.vstack(list(gen))
        expected_data = np.load(os.path.join(
            dirpath, 'protochips_gas_cell.npy'))
        np.testing.assert_allclose(data.T, expected_data)

    def test_read_metadata_header(self):
        assert self.pgc.time_units == 'Milliseconds'
        assert self.pgc.time_units == 'Milliseconds'
        assert self.pgc.temperature_units == 'Degrees C'
        assert self.pgc.pressure_units == 'Torr'
        assert self.pgc.current_units == 'Amps'
        assert self.pgc.voltage_units == 'Volts'
        assert self.pgc.resistance_units == 'Ohms'
        assert self.pgc.user == 'eric'


#########################
# Protochips electrical #
#########################

def test_read_protochips_electrical():
    filename = os.path.join(dirpath, 'protochips_electrical.csv')
    s = hs.load(filename)
    assert len(s) == 6
    assert s[0].metadata.General.title == 'Channel A Current (Amps)'
    assert s[0].metadata.Signal.signal_type == ''
    assert s[0].metadata.Signal.quantity == 'Current (Amps)'
    assert s[1].metadata.General.title == 'Channel A Voltage (Volts)'
    assert s[1].metadata.Signal.signal_type == ''
    assert s[1].metadata.Signal.quantity == 'Voltage (Volts)'
    assert s[2].metadata.General.title == 'Channel A Resistance (Ohms)'
    assert s[2].metadata.Signal.signal_type == ''
    assert s[2].metadata.Signal.quantity == 'Resistance (Ohms)'
    assert s[3].metadata.General.title == 'Channel B Current (Amps)'
    assert s[3].metadata.Signal.signal_type == ''
    assert s[3].metadata.Signal.quantity == 'Current (Amps)'
    assert s[4].metadata.General.title == 'Channel B Voltage (Volts)'
    assert s[4].metadata.Signal.signal_type == ''
    assert s[4].metadata.Signal.quantity == 'Voltage (Volts)'
    assert s[5].metadata.General.title == 'Channel B Resistance (Ohms)'
    assert s[5].metadata.Signal.signal_type == ''
    assert s[5].metadata.Signal.quantity == 'Resistance (Ohms)'


class TestProtochipsElectricalCSVReader():

    def setup_method(self, method):
        self.filename = os.path.join(dirpath, 'protochips_electrical.csv')
        self.pa = ProtochipsCSV(self.filename)
        if generate_numpy_file:
            create_numpy_file(self.filename.replace('.csv', '.npy'), self.pa)

    def test_read_column_name(self):
        assert self.pa.column_name == ['Time', 'Notes',
                                       'Channel A Current',
                                       'Channel A Voltage',
                                       'Channel A Resistance',
                                       'Channel B Current',
                                       'Channel B Voltage',
                                       'Channel B Resistance']

    def test_read_start_datetime(self):
        dt = np.datetime64('2014-10-08T16:26:51.738000')
        assert self.pa.start_datetime == dt

    def test_read_data(self):
        gen = (self.pa._data_dictionary[key]
               for key in self.pa.logged_quantity_name_list)
        data = np.vstack(list(gen))
        expected_data = np.load(os.path.join(
            dirpath, 'protochips_electrical.npy'))
        np.testing.assert_allclose(data.T, expected_data)


######################
# Protochips thermal #
######################

def test_read_protochips_thermal():
    filename = os.path.join(dirpath, 'protochips_thermal.csv')
    s = hs.load(filename)
    assert (s.metadata.General.title ==
            'Channel A Temperature (Degrees C)')
    assert s.metadata.Signal.signal_type == ''
    assert s.metadata.Signal.quantity == 'Temperature (Degrees C)'
    assert (s.metadata.General.notes ==
            'Calibration file name: AD21013_8.cal')


class TestProtochipsThermallCSVReader():

    def setup_method(self, method):
        self.filename = os.path.join(dirpath, 'protochips_thermal.csv')
        self.pt = ProtochipsCSV(self.filename)
        if generate_numpy_file:
            create_numpy_file(self.filename.replace('.csv', '.npy'), self.pt)

    def test_read_column_name(self):
        assert self.pt.column_name == [
            'Time', 'Notes', 'Channel A Temperature']

    def test_read_start_datetime(self):
        dt = np.datetime64('2014-12-03T17:15:37.192000')
        np.testing.assert_equal(self.pt.start_datetime, dt)

    def test_read_data(self):
        gen = (self.pt._data_dictionary[key]
               for key in self.pt.logged_quantity_name_list)
        data = np.vstack(list(gen))
        expected_data = np.load(os.path.join(
            dirpath, 'protochips_thermal.npy'))
        np.testing.assert_allclose(data.T, expected_data)


#############################
# Protochips electrothermal #
#############################

def test_read_protochips_electrothermal():
    filename = os.path.join(dirpath, 'protochips_electrothermal.csv')
    s = hs.load(filename)
    assert len(s) == 4
    assert (s[0].metadata.General.title ==
            'Channel A Temperature (Degrees C)')
    assert s[0].metadata.Signal.signal_type == ''
    assert s[0].metadata.Signal.quantity == 'Temperature (Degrees C)'
    assert s[1].metadata.General.title == 'Channel B Current (Amps)'
    assert s[1].metadata.Signal.signal_type == ''
    assert (s[0].metadata.General.notes ==
            'Calibration file name: AD21018_4.cal')
    assert s[1].metadata.Signal.quantity == 'Current (Amps)'
    assert s[2].metadata.General.title == 'Channel B Voltage (Volts)'
    assert s[2].metadata.Signal.signal_type == ''
    assert s[2].metadata.Signal.quantity == 'Voltage (Volts)'
    assert s[3].metadata.General.title == 'Channel B Resistance (Ohms)'
    assert s[3].metadata.Signal.signal_type == ''
    assert s[3].metadata.Signal.quantity == 'Resistance (Ohms)'


class TestProtochipsElectrothermalCSVReader():

    def setup_method(self, method):
        self.filename = os.path.join(dirpath, 'protochips_electrothermal.csv')
        self.pet = ProtochipsCSV(self.filename)
        if generate_numpy_file:
            create_numpy_file(self.filename.replace('.csv', '.npy'), self.pet)

    def test_read_column_name(self):
        assert self.pet.column_name == ['Time', 'Notes',
                                        'Channel A Temperature',
                                        'Channel B Current',
                                        'Channel B Voltage',
                                        'Channel B Resistance']

    def test_read_start_datetime(self):
        dt = np.datetime64('2014-11-05T14:42:51.369000')
        np.testing.assert_equal(self.pet.start_datetime, dt)

    def test_read_data(self):
        gen = (self.pet._data_dictionary[key]
               for key in self.pet.logged_quantity_name_list)
        data = np.vstack(list(gen))
        expected_data = np.load(os.path.join(dirpath,
                                             self.filename.replace('.csv', '.npy')))
        np.testing.assert_allclose(data.T, expected_data)



# -*- coding: utf-8 -*-
# Copyright 2007-2020 The HyperSpy developers
#
# This file is part of  HyperSpy.
#
#  HyperSpy is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
#  HyperSpy is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with  HyperSpy.  If not, see <http://www.gnu.org/licenses/>.

import logging
import warnings
from itertools import chain

import numpy as np
import scipy.linalg

from hyperspy.exceptions import VisibleDeprecationWarning
from hyperspy.external.progressbar import progressbar
from hyperspy.learn.svd_pca import svd_solve
from hyperspy.misc.math_tools import check_random_state

_logger = logging.getLogger(__name__)


def _soft_thresh(X, lambda1):
    """Soft-thresholding of array X."""
    res = np.abs(X) - lambda1
    np.maximum(res, 0.0, out=res)
    res *= np.sign(X)
    return res


def rpca_godec(
    X, rank, lambda1=None, power=0, tol=1e-3, maxiter=1000, random_state=None, **kwargs
):
    """Perform Robust PCA with missing or corrupted data, using the GoDec algorithm.

    Decomposes a matrix Y = X + E, where X is low-rank and E
    is a sparse error matrix. This algorithm is based on the
    Matlab code from [Zhou2011]_. See code here:
    https://sites.google.com/site/godecomposition/matrix/artifact-1

    Read more in the :ref:`User Guide <mva.rpca>`.

    Parameters
    ----------
    X : numpy array, shape (n_features, n_samples)
        The matrix of observations.
    rank : int
        The model dimensionality.
    lambda1 : None or float
        Regularization parameter.
        If None, set to 1 / sqrt(n_features)
    power : int, default 0
        The number of power iterations used in the initialization
    tol : float, default 1e-3
        Convergence tolerance
    maxiter : int, default 1000
        Maximum number of iterations
    random_state : None or int or RandomState instance, default None
        Used to initialize the subspace on the first iteration.

    Returns
    -------
    Xhat : numpy array, shape (n_features, n_samples)
        The  low-rank matrix
    Ehat : numpy array, shape (n_features, n_samples)
        The sparse error matrix
    U, S, V : numpy arrays
        The results of an SVD on Xhat

    References
    ----------
    .. [Zhou2011] Tianyi Zhou and Dacheng Tao, "GoDec: Randomized Low-rank &
        Sparse Matrix Decomposition in Noisy Case", ICML-11, (2011), pp. 33-40.

    """
    # Get shape
    m, n = X.shape

    # Operate on transposed matrix for speed
    transpose = False
    if m < n:
        transpose = True
        X = X.T

    # Get shape
    m, n = X.shape

    # Check options if None
    if lambda1 is None:
        _logger.info("Threshold 'lambda1' is set to default: 1 / sqrt(n_features)")
        lambda1 = 1.0 / np.sqrt(m)

    # Initialize L and E
    L = X
    E = np.zeros(L.shape)

    random_state = check_random_state(random_state)

    for itr in range(int(maxiter)):
        # Initialization with bilateral random projections
        Y2 = random_state.normal(size=(n, rank))
        for _ in range(power + 1):
            Y2 = L.T @ (L @ Y2)

        Q, _ = scipy.linalg.qr(Y2, mode="economic")

        # Estimate the new low-rank and sparse matrices
        Lnew = (L @ Q) @ Q.T
        A = L - Lnew + E
        L = Lnew
        E = _soft_thresh(A, lambda1)
        A -= E
        L += A

        # Check convergence
        eps = np.linalg.norm(A)
        if eps < tol:
            _logger.info(f"Converged to {eps} in {itr} iterations")
            break

    # Transpose back
    if transpose:
        L = L.T
        E = E.T

    # Rescale
    Xhat = L
    Ehat = E

    # Do final SVD
    U, S, Vh = svd_solve(Xhat, output_dimension=rank, **kwargs)
    V = Vh.T

    # Chop small singular values which
    # likely arise from numerical noise
    # in the SVD.
    S[rank:] = 0.0

    return Xhat, Ehat, U, S, V


def _solveproj(z, X, Id, lambda2, r=None, e=None):
    m, n = X.shape
    z = z.T

    if len(z.shape) == 2:
        batch_size = z.shape[1]
        eshape = (m, batch_size)
        rshape = (n, batch_size)
    else:
        eshape = (m,)
        rshape = (n,)
    if r is None or r.shape != rshape:
        r = np.zeros(rshape)
    if e is None or e.shape != eshape:
        e = np.zeros(eshape)

    ddt = np.linalg.solve(X.T @ X + Id, X.T)
    maxiter = 1e6
    itr = 0

    while True:
        itr += 1
        # Solve for r
        rtmp = r
        r = ddt @ (z - e)

        # Solve for e
        etmp = e
        e = _soft_thresh(z - X @ r, lambda2)

        # Stop conditions
        stopr = np.linalg.norm(r - rtmp, 2)
        stope = np.linalg.norm(e - etmp, 2)
        stop = max(stopr, stope) / m
        if stop < 1e-5 or itr > maxiter:
            break

    return r, e


def _updatecol(X, A, B, Id):
    tmp, n = X.shape
    L = X
    A = A + Id

    for i in range(n):
        b = B[:, i]
        x = X[:, i]
        a = A[:, i]
        temp = (b - X @ a) / A[i, i] + x
        L[:, i] = temp / max(np.linalg.norm(temp, 2), 1)

    return L


class ORPCA:
    """Performs Online Robust PCA with missing or corrupted data.

    The ORPCA code is based on a transcription of MATLAB code from [Feng2013]_.
    It has been updated to include a new initialization method based
    on a QR decomposition of the first n "training" samples of the data.
    A stochastic gradient descent (SGD) solver is also implemented,
    along with a MomentumSGD solver for improved convergence and robustness
    with respect to local minima. More information about the gradient descent
    methods and choosing appropriate parameters can be found in [Ruder2016]_.

    Read more in the :ref:`User Guide <mva.rpca>`.

    References
    ----------
    .. [Feng2013] Jiashi Feng, Huan Xu and Shuicheng Yuan, "Online Robust PCA
        via Stochastic Optimization", Advances in Neural Information Processing
        Systems 26, (2013), pp. 404-412.
    .. [Ruder2016] Sebastian Ruder, "An overview of gradient descent optimization
        algorithms", arXiv:1609.04747, (2016), http://arxiv.org/abs/1609.04747.

    """

    def __init__(
        self,
        rank,
        store_error=False,
        lambda1=0.1,
        lambda2=1.0,
        method="BCD",
        init="qr",
        training_samples=10,
        subspace_learning_rate=1.0,
        subspace_momentum=0.5,
        random_state=None,
    ):
        """Creates Online Robust PCA instance that can learn a representation.

        Parameters
        ----------
        rank : int
            The rank of the representation (number of components/factors)
        store_error : bool, default False
            If True, stores the sparse error matrix.
        lambda1 : float
            Nuclear norm regularization parameter.
        lambda2 : float
            Sparse error regularization parameter.
        method : {'CF', 'BCD', 'SGD', 'MomentumSGD'}, default 'BCD'
            * 'CF'  - Closed-form solver
            * 'BCD' - Block-coordinate descent
            * 'SGD' - Stochastic gradient descent
            * 'MomentumSGD' - Stochastic gradient descent with momentum
        init : {'qr', 'rand', np.ndarray}, default 'qr'
            * 'qr'   - QR-based initialization
            * 'rand' - Random initialization
            * np.ndarray if the shape [n_features x rank]
        training_samples : int
            Specifies the number of training samples to use in
            the 'qr' initialization.
        subspace_learning_rate : float
            Learning rate for the 'SGD' and 'MomentumSGD' methods.
            Should be a float > 0.0
        subspace_momentum : float
            Momentum parameter for 'MomentumSGD' method, should be
            a float between 0 and 1.
        random_state : None or int or RandomState instance, default None
            Used to initialize the subspace on the first iteration.

        """
        self.n_features = None
        self.iterating = False
        self.t = 0

        if store_error:
            self.E = []
        else:
            self.E = None

        self.rank = rank
        self.lambda1 = lambda1
        self.lambda2 = lambda2
        self.method = method
        self.init = init
        self.training_samples = training_samples
        self.subspace_learning_rate = subspace_learning_rate
        self.subspace_momentum = subspace_momentum
        self.random_state = check_random_state(random_state)

        # Check options are valid
        if method not in ("CF", "BCD", "SGD", "MomentumSGD"):
            raise ValueError("'method' not recognised")
        if not isinstance(init, np.ndarray) and init not in ("qr", "rand"):
            raise ValueError("'init' not recognised")
        if not isinstance(init, np.ndarray):
            if init == "qr" and training_samples < rank:
                raise ValueError("'training_samples' must be >= 'output_dimension'")
        if method == "MomentumSGD" and (
            subspace_momentum > 1.0 or subspace_momentum < 0.0
        ):
            raise ValueError("'subspace_momentum' must be a float between 0 and 1")

    def _setup(self, X):
        self.r, self.e, self.v = None, None, None
        if isinstance(X, np.ndarray):
            n, m = X.shape
            iterating = False
        else:
            x = next(X)
            m = len(x)
            X = chain([x], X)
            iterating = True

        self.n_features = m
        self.iterating = iterating

        self.L = self._initialize_subspace(X)
        self.K = self.lambda1 * np.eye(self.rank)
        self.R = []

        # Extra variables for CF and BCD methods
        if self.method in ("CF", "BCD"):
            self.A = np.zeros((self.rank, self.rank))
            self.B = np.zeros((m, self.rank))
        elif self.method == "MomentumSGD":
            self.vnew = np.zeros_like(self.L)

        return X

    def _initialize_subspace(self, X):
        """Initialize the subspace estimate."""
        m = self.n_features

        if isinstance(self.init, np.ndarray):
            if self.init.ndim != 2:
                raise ValueError("'init' has to be a two-dimensional matrix")
            init_m, init_r = self.init.shape
            if init_m != m or init_r != self.rank:
                raise ValueError("'init' has to be of shape [n_features x rank]")
            return self.init.copy()
        elif self.init == "qr":
            if self.iterating:
                Y2 = np.stack([next(X) for _ in range(self.training_samples)], axis=-1)
                X = chain(iter(Y2.T.copy()), X)
            else:
                Y2 = X[: self.training_samples, :].T
            L, _ = scipy.linalg.qr(Y2, mode="economic")
            return L[:, : self.rank]
        elif self.init == "rand":
            Y2 = self.random_state.normal(size=(m, self.rank))
            L, _ = scipy.linalg.qr(Y2, mode="economic")
            return L[:, : self.rank]

    def fit(self, X, batch_size=None):
        """Learn RPCA components from the data.

        Parameters
        ----------
        X : {numpy.ndarray, iterator}
            [n_samples x n_features] matrix of observations
            or an iterator that yields samples, each with n_features elements.
        batch_size : {None, int}
            If not None, learn the data in batches, each of batch_size samples
            or less.

        """
        if self.n_features is None:
            X = self._setup(X)

        num = None
        prod = np.outer
        if batch_size is not None:
            if not isinstance(X, np.ndarray):
                raise ValueError("can't batch iterating data")
            else:
                prod = np.dot
                length = X.shape[0]
                num = max(length // batch_size, 1)
                X = np.array_split(X, num, axis=0)

        if isinstance(X, np.ndarray):
            num = X.shape[0]
            X = iter(X)

        r, e = self.r, self.e

        for v in progressbar(X, leave=False, total=num, disable=num == 1):
            r, e = _solveproj(v, self.L, self.K, self.lambda2, r=r, e=e)
            self.v = v
            self.r = r
            self.e = e
            self.R.append(r)
            if self.E is not None:
                self.E.append(e)

            self._solve_L(prod(r, r.T), prod((v.T - e), r.T))
            self.t += 1

        self.r = r
        self.e = e

    def _solve_L(self, A, B):
        if self.method == "CF":
            # Closed-form solution
            self.A += A
            self.B += B
            self.L = self.B @ np.linalg.inv(self.A + self.K)
        elif self.method == "BCD":
            # Block-coordinate descent
            self.A += A
            self.B += B
            self.L = _updatecol(self.L, self.A, self.B, self.K)
        elif self.method == "SGD":
            # Stochastic gradient descent
            learn = self.subspace_learning_rate * (
                1 + self.subspace_learning_rate * self.lambda1 * self.t
            )
            self.L -= (self.L @ A - B + self.lambda1 * self.L) / learn
        elif self.method == "MomentumSGD":
            # Stochastic gradient descent with momentum
            learn = self.subspace_learning_rate * (
                1 + self.subspace_learning_rate * self.lambda1 * self.t
            )
            vold = self.subspace_momentum * self.vnew
            self.vnew = (self.L @ A - B + self.lambda1 * self.L) / learn
            self.L -= vold + self.vnew

    def project(self, X, return_error=False):
        """Project the learnt components on the data.

        Parameters
        ----------
        X : {numpy.ndarray, iterator}
            [n_samples x n_features] matrix of observations
            or an iterator that yields n_samples, each with n_features elements.
        return_error : bool
            If True, returns the sparse error matrix as well. Otherwise only
            the weights (loadings)

        """
        R = []
        if return_error:
            E = []

        num = None
        if isinstance(X, np.ndarray):
            num = X.shape[0]
            X = iter(X)
        for v in progressbar(X, leave=False, total=num):
            r, e = _solveproj(v, self.L, self.K, self.lambda2)
            R.append(r.copy())
            if return_error:
                E.append(e.copy())

        R = np.stack(R, axis=-1)
        if return_error:
            return R, np.stack(E, axis=-1)
        else:
            return R

    def finish(self, **kwargs):
        """Return the learnt factors and loadings."""
        if len(self.R) > 0:
            if len(self.R[0].shape) == 1:
                R = np.stack(self.R, axis=-1)
            else:
                R = np.concatenate(self.R, axis=1)
            return self.L, R
        else:
            return self.L, 1


def orpca(
    X,
    rank,
    store_error=False,
    project=False,
    batch_size=None,
    lambda1=0.1,
    lambda2=1.0,
    method="BCD",
    init="qr",
    training_samples=10,
    subspace_learning_rate=1.0,
    subspace_momentum=0.5,
    random_state=None,
    **kwargs,
):
    """Perform online, robust PCA on the data X.

    This is a wrapper function for the ORPCA class.

    Parameters
    ----------
    X : {numpy array, iterator}
        [n_features x n_samples] matrix of observations
        or an iterator that yields samples, each with n_features elements.
    rank : int
        The rank of the representation (number of components/factors)
    store_error : bool, default False
        If True, stores the sparse error matrix.
    project : bool, default False
        If True, project the data X onto the learnt model.
    batch_size : {None, int}, default None
        If not None, learn the data in batches, each of batch_size samples
        or less.
    lambda1 : float
        Nuclear norm regularization parameter.
    lambda2 : float
        Sparse error regularization parameter.
    method : {'CF', 'BCD', 'SGD', 'MomentumSGD'}, default 'BCD'
        * 'CF'  - Closed-form solver
        * 'BCD' - Block-coordinate descent
        * 'SGD' - Stochastic gradient descent
        * 'MomentumSGD' - Stochastic gradient descent with momentum
    init : {'qr', 'rand', np.ndarray}, default 'qr'
        * 'qr'   - QR-based initialization
        * 'rand' - Random initialization
        * np.ndarray if the shape [n_features x rank]
    training_samples : int
        Specifies the number of training samples to use in
        the 'qr' initialization.
    subspace_learning_rate : float
        Learning rate for the 'SGD' and 'MomentumSGD' methods.
        Should be a float > 0.0
    subspace_momentum : float
        Momentum parameter for 'MomentumSGD' method, should be
        a float between 0 and 1.
    random_state : None or int or RandomState instance, default None
        Used to initialize the subspace on the first iteration.

    Returns
    -------
    numpy arrays
        * If project is True, returns the low-rank factors and loadings only
        * Otherwise, returns the low-rank and sparse error matrices, as well
          as the results of a singular value decomposition (SVD) applied to
          the low-rank matrix.

    """
    if kwargs.get("learning_rate", False):
        warnings.warn(
            "The argument `learning_rate` has been deprecated and may "
            "be removed in future. Please use `subspace_learning_rate` instead.",
            VisibleDeprecationWarning,
        )
        subspace_learning_rate = kwargs["learning_rate"]
    if kwargs.get("momentum", False):
        warnings.warn(
            "The argument `momentum` has been deprecated and may "
            "be removed in future. Please use `subspace_momentum` instead.",
            VisibleDeprecationWarning,
        )
        subspace_momentum = kwargs["momentum"]

    X = X.T

    _orpca = ORPCA(
        rank,
        store_error=store_error,
        lambda1=lambda1,
        lambda2=lambda2,
        method=method,
        init=init,
        training_samples=training_samples,
        subspace_learning_rate=subspace_learning_rate,
        subspace_momentum=subspace_momentum,
        random_state=random_state,
    )
    _orpca.fit(X, batch_size=batch_size)

    if project:
        L = _orpca.L
        R = _orpca.project(X)
    else:
        L, R = _orpca.finish()

    if store_error:
        Xhat = L @ R
        Ehat = np.array(_orpca.E).T

        # Do final SVD
        U, S, Vh = svd_solve(Xhat, output_dimension=rank)
        V = Vh.T

        # Chop small singular values which
        # likely arise from numerical noise
        # in the SVD.
        S[rank:] = 0.0

        return Xhat, Ehat, U, S, V
    else:
        return L, R



import copy
import os.path
import tempfile

from hyperspy.io import load
from hyperspy.misc.test_utils import assert_deep_almost_equal

my_path = os.path.dirname(__file__)

example1_TEM = {'Detector': {'EELS': {'collection_angle': 3.4,
                                      'collection_angle_units': "mR",
                                      'dwell_time': 100.0,
                                      'dwell_time_units': "ms"}},
                'beam_current': 12.345,
                'beam_current_units': "nA",
                'beam_energy': 120.0,
                'beam_energy_units': "kV",
                'convergence_angle': 1.5,
                'convergence_angle_units': "mR"}

example1_metadata = {'Acquisition_instrument': {'TEM': example1_TEM},
                     'General': {'original_filename': "example1.msa",
                                 'title': "NIO EELS OK SHELL",
                                 'date': "1991-10-01",
                                 'time': "12:00:00"},
                     'Sample': {'thickness': 50.0,
                                'thickness_units': "nm"},
                     'Signal': {'binned': True,
                                # bit silly...
                                'quantity': "Counts (Intensity)",
                                'signal_type': 'EELS'},
                     '_HyperSpy': {'Folding': {'original_axes_manager': None,
                                               'unfolded': False,
                                               'original_shape': None,
                                               'signal_unfolded': False}}}
minimum_md_om = {
    'COMMENT': 'File created by HyperSpy version 1.1.2+dev',
    'DATATYPE': 'Y',
    'DATE': '',
    'FORMAT': 'EMSA/MAS Spectral Data File',
    'NCOLUMNS': 1.0,
    'NPOINTS': 1.0,
    'OFFSET': 0.0,
    'OWNER': '',
    'SIGNALTYPE': '',
    'TIME': '',
    'TITLE': '',
    'VERSION': '1.0',
    'XLABEL': '',
    'XPERCHAN': 1.0,
    'XUNITS': ''}


example1_parameters = {
    'BEAMDIAM -nm': 100.0,
    'BEAMKV   -kV': 120.0,
    'CHOFFSET': -168.0,
    'COLLANGLE-mR': 3.4,
    'CONVANGLE-mR': 1.5,
    'DATATYPE': 'XY',
    'DATE': '01-OCT-1991',
    'DWELLTIME-ms': 100.0,
    'ELSDET': 'SERIAL',
    'EMISSION -uA': 5.5,
    'FORMAT': 'EMSA/MAS Spectral Data File',
    'MAGCAM': 100.0,
    'NCOLUMNS': 1.0,
    'NPOINTS': 20.0,
    'OFFSET': 520.13,
    'OPERMODE': 'IMAG',
    'OWNER': 'EMSA/MAS TASK FORCE',
    'PROBECUR -nA': 12.345,
    'SIGNALTYPE': 'ELS',
    'THICKNESS-nm': 50.0,
    'TIME': '12:00',
    'TITLE': 'NIO EELS OK SHELL',
    'VERSION': '1.0',
    'XLABEL': 'Energy',
    'XPERCHAN': 3.1,
    'XUNITS': 'eV',
    'YLABEL': 'Counts',
    'YUNITS': 'Intensity'}

example2_TEM = {'Detector': {'EDS': {'EDS_det': "SIWLS",
                                     'azimuth_angle': 90.0,
                                     'azimuth_angle_units': "dg",
                                     'elevation_angle': 20.0,
                                     'elevation_angle_units': 'dg',
                                     'live_time': 100.0,
                                     'live_time_units': "s",
                                     'real_time': 150.0,
                                     'real_time_units': "s"}},
                'beam_current': 12.345,
                'beam_current_units': "nA",
                'beam_energy': 120.0,
                'beam_energy_units': "kV",
                'Stage': {'tilt_alpha': 45.0,
                          'tilt_alpha_units': "dg"}}

example2_metadata = {'Acquisition_instrument': {'TEM': example2_TEM},
                     'General': {'original_filename': "example2.msa",
                                 'title': "NIO Windowless Spectra OK NiL",
                                 'date': "1991-10-01",
                                 'time': "12:00:00"},
                     'Sample': {'thickness': 50.0,
                                'thickness_units': "nm"},
                     'Signal': {'binned': False,
                                'quantity': "X-RAY INTENSITY (Intensity)",
                                'signal_type': 'EDS'},
                     '_HyperSpy': {'Folding': {'original_axes_manager': None,
                                               'unfolded': False,
                                               'original_shape': None,
                                               'signal_unfolded': False}}}

example2_parameters = {
    'ALPHA-1': '3.1415926535',
    'AZIMANGLE-dg': 90.0,
    'BEAMDIAM -nm': 100.0,
    'BEAMKV   -kV': 120.0,
    'CHOFFSET': -20.0,
    'COMMENT': 'The next two lines are User Defined Keywords and values',
    'DATATYPE': 'Y',
    'DATE': '01-OCT-1991',
    'EDSDET': 'SIWLS',
    'ELEVANGLE-dg': 20.0,
    'EMISSION -uA': 5.5,
    'FORMAT': 'EMSA/MAS Spectral Data File',
    'LIVETIME  -s': 100.0,
    'MAGCAM': 100.0,
    'NCOLUMNS': 5.0,
    'NPOINTS': 80.0,
    'OFFSET': 200.0,
    'OPERMODE': 'IMAG',
    'OWNER': 'EMSA/MAS TASK FORCE',
    'PROBECUR -nA': 12.345,
    'REALTIME  -s': 150.0,
    'RESTMASS': '511.030',
    'SIGNALTYPE': 'EDS',
    'SOLIDANGL-sR': '0.13',
    'TACTLYR  -cm': 0.3,
    'TAUWIND  -cm': 2e-06,
    'TBEWIND  -cm': 0.0,
    'TDEADLYR -cm': 1e-06,
    'THICKNESS-nm': 50.0,
    'TIME': '12:00',
    'TITLE': 'NIO Windowless Spectra OK NiL',
    'VERSION': '1.0',
    'XLABEL': 'X-RAY ENERGY',
    'XPERCHAN': 10.0,
    'XPOSITION': 123.0,
    'XTILTSTGE-dg': 45.0,
    'XUNITS': 'eV',
    'YLABEL': 'X-RAY INTENSITY',
    'YPOSITION': 456.0,
    'YTILTSTGE-dg': 20.0,
    'YUNITS': 'Intensity',
    'ZPOSITION': 0.0}


class TestExample1:

    def setup_method(self, method):
        self.s = load(os.path.join(
            my_path,
            "msa_files",
            "example1.msa"))

    def test_data(self):
        assert (
            [4066.0,
             3996.0,
             3932.0,
             3923.0,
             5602.0,
             5288.0,
             7234.0,
             7809.0,
             4710.0,
             5015.0,
             4366.0,
             4524.0,
             4832.0,
             5474.0,
             5718.0,
             5034.0,
             4651.0,
             4613.0,
             4637.0,
             4429.0,
             4217.0] == self.s.data.tolist())

    def test_parameters(self):
        assert (
            example1_parameters ==
            self.s.original_metadata.as_dictionary())

    def test_metadata(self):
        assert_deep_almost_equal(self.s.metadata.as_dictionary(),
                                 example1_metadata)

    def test_write_load_cycle(self):
        with tempfile.TemporaryDirectory() as tmpdir:
            fname2 = os.path.join(tmpdir, "example1-export.msa")
            self.s.save(fname2)
            s2 = load(fname2)
            assert (s2.metadata.General.original_filename ==
                    "example1-export.msa")
            s2.metadata.General.original_filename = "example1.msa"
            assert_deep_almost_equal(self.s.metadata.as_dictionary(),
                                     s2.metadata.as_dictionary())

class TestExample1WrongDate:

    def setup_method(self, method):
        self.s = load(os.path.join(
            my_path,
            "msa_files",
            "example1_wrong_date.msa"))

    def test_metadata(self):
        md = copy.copy(example1_metadata)
        del md["General"]["date"]
        del md["General"]["time"]
        md["General"]["original_filename"] = "example1_wrong_date.msa"
        assert_deep_almost_equal(self.s.metadata.as_dictionary(),
                                 md)




class TestExample2:

    def setup_method(self, method):
        self.s = load(os.path.join(
            my_path,
            "msa_files",
            "example2.msa"))

    def test_data(self):
        assert (
            [65.82,
             67.872,
             65.626,
             68.762,
             71.395,
             74.996,
             78.132,
             78.055,
             77.861,
             84.598,
             83.088,
             85.372,
             89.786,
             93.464,
             93.387,
             97.452,
             109.96,
             111.08,
             119.64,
             128.77,
             138.38,
             152.35,
             176.01,
             192.12,
             222.12,
             254.22,
             281.55,
             328.33,
             348.92,
             375.33,
             385.51,
             389.54,
             378.77,
             353.8,
             328.91,
             290.07,
             246.09,
             202.73,
             176.47,
             137.64,
             119.56,
             106.4,
             92.496,
             96.213,
             94.664,
             101.13,
             114.57,
             118.82,
             131.68,
             145.04,
             165.44,
             187.51,
             207.49,
             238.04,
             269.71,
             301.46,
             348.65,
             409.36,
             475.3,
             554.51,
             631.64,
             715.19,
             793.44,
             847.99,
             872.97,
             862.59,
             834.87,
             778.49,
             688.63,
             599.39,
             495.39,
             403.48,
             312.88,
             237.34,
             184.14,
             129.86,
             101.59,
             80.107,
             58.657,
             49.442] == self.s.data.tolist())

    def test_parameters(self):
        assert (
            example2_parameters ==
            self.s.original_metadata.as_dictionary())

    def test_metadata(self):
        assert_deep_almost_equal(self.s.metadata.as_dictionary(),
                                 example2_metadata)

    def test_write_load_cycle(self):
        with tempfile.TemporaryDirectory() as tmpdir:
            fname2 = os.path.join(tmpdir, "example2-export.msa")
            self.s.save(fname2)
            s2 = load(fname2)
            assert (s2.metadata.General.original_filename ==
                    "example2-export.msa")
            s2.metadata.General.original_filename = "example2.msa"
            assert_deep_almost_equal(self.s.metadata.as_dictionary(),
                                     s2.metadata.as_dictionary())


def test_minimum_metadata_example():
    s = load(os.path.join(my_path, "msa_files", "minimum_metadata.msa"))
    assert minimum_md_om == s.original_metadata.as_dictionary()



# -*- coding: utf-8 -*-
# Copyright 2007-2020 The HyperSpy developers
#
# This file is part of  HyperSpy.
#
#  HyperSpy is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
#  HyperSpy is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with  HyperSpy.  If not, see <http://www.gnu.org/licenses/>.

import pytest
import numpy as np

from hyperspy.components1d import Arctan
from hyperspy.exceptions import VisibleDeprecationWarning


def test_function():
    g = Arctan()
    g.A.value = 10
    g.k.value = 2
    g.x0.value = 1
    np.testing.assert_allclose(g.function(0), -11.07148718)
    np.testing.assert_allclose(g.function(1), 0)
    np.testing.assert_allclose(g.function(1e4), 10*np.pi/2,1e-4)

# Legacy tests
def test_function_legacyF():
    g = Arctan(minimum_at_zero=False)
    g.A.value = 10
    g.k.value = 2
    g.x0.value = 1
    np.testing.assert_allclose(g.function(0), -11.07148718)
    np.testing.assert_allclose(g.function(1), 0)
    np.testing.assert_allclose(g.function(1e4), 10*np.pi/2,1e-4)

def test_function_legacyT():
    with pytest.warns(
            VisibleDeprecationWarning,
            match="component will change in v2.0.",
        ):
        g = Arctan(minimum_at_zero=True)

    g.A.value = 10
    g.k.value = 2
    g.x0.value = 1
    np.testing.assert_allclose(g.function(0), 4.63647609)
    np.testing.assert_allclose(g.function(1), 10*np.pi/2)
    np.testing.assert_allclose(g.function(1e4), 10*np.pi,1e-4)



# -*- coding: utf-8 -*-
# Copyright 2007-2020 The HyperSpy developers
#
# This file is part of  HyperSpy.
#
#  HyperSpy is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
#  HyperSpy is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with  HyperSpy.  If not, see <http://www.gnu.org/licenses/>.

import numpy as np

from hyperspy.component import Component
from hyperspy._components.gaussian import _estimate_gaussian_parameters
from hyperspy.docstrings.parameters import FUNCTION_ND_DOCSTRING


sqrt2pi = np.sqrt(2 * np.pi)


class SplitVoigt(Component):

    r"""Split pseudo-Voigt

    .. math::
        :nowrap:

        \[
        pV(x,centre,\sigma) = (1 - \eta) G(x,centre,\sigma)
        + \eta L(x,centre,\sigma)
        \]


        \[
        f(x) =
        \begin{cases}
            pV(x,centre,\sigma_1), & x \leq centre\\
            pV(x,centre,\sigma_2), & x >  centre
        \end{cases}
        \]

    ================= ===========
    Variable           Parameter
    ================= ===========
    :math:`A`          A
    :math:`\eta`       fraction
    :math:`\sigma_1`   sigma1
    :math:`\sigma_2`   sigma2
    :math:`centre`     centre
    ================= ===========

    Note
    -----
    This is a voigt function in which the upstream and downstream variance or
    sigma is allowed to vary to create an asymmetric profile
    In this case the voigt is a pseudo voigt- consisting of a
    mixed gaussian and lorentzian sum

    """

    def __init__(self, A=1., sigma1=1., sigma2=1.0, fraction=0.0, centre=0.):
        Component.__init__(
            self, ('A', 'sigma1', 'sigma2', 'centre', 'fraction'))
        self.A.value = A
        self.sigma1.value = sigma1
        self.sigma2.value = sigma2
        self.centre.value = centre
        self.fraction.value = fraction

        # Boundaries
        self.A.bmin = 1.0e-8
        self.A.bmax = 1e8
        self.sigma2.bmin = 1.0e-8
        self.sigma2.bmax = 50.0
        self.sigma1.bmin = 1.0e-8
        self.sigma1.bmax = 50.0
        self.fraction.bmin = 1.0e-8
        self.fraction.bmax = 1.0
        self.isbackground = False
        self.convolved = True

    def _function(self, x, A, sigma1, sigma2, fraction, centre):
        arg = (x - centre)
        lor1 = (A / (1.0 + ((1.0 * arg) / sigma1) ** 2)) \
            / (0.5 * np.pi * (sigma1 + sigma2))
        lor2 = (A / (1.0 + ((1.0 * arg) / sigma2) ** 2)) \
            / (0.5 * np.pi * (sigma1 + sigma2))

        prefactor = A / (sqrt2pi * 0.5 * (sigma1 + sigma2))
        gauss1 = prefactor * np.exp(-0.5 * arg * arg / (sigma1 * sigma1))
        gauss2 = prefactor * np.exp(-0.5 * arg * arg / (sigma2 * sigma2))

        p1 = (1.0 - fraction) * gauss1 + fraction * lor1
        p2 = (1.0 - fraction) * gauss2 + fraction * lor2

        return np.where(x <= centre, p1, p2)

    def function(self, x):
        """Split pseudo voigt - a linear combination  of gaussian and lorentzian

        Parameters
        ----------
        x : array
            independent variable
        A : float
            area of pvoigt peak
        center : float
            center position
        sigma1 : float
            standard deviation <= center position
        sigma2 : float
            standard deviation > center position
        fraction : float
            weight for lorentzian peak in the linear combination,
            and (1-fraction) is the weight for gaussian peak.
        """
        A = self.A.value
        sigma1 = self.sigma1.value
        sigma2 = self.sigma2.value
        fraction = self.fraction.value
        centre = self.centre.value

        return self._function(x, A, sigma1, sigma2, fraction, centre)

    def function_nd(self, axis):
        """%s

        """
        if self._is_navigation_multidimensional:
            x = axis[np.newaxis, :]
            A = self.A.map['values'][..., np.newaxis]
            sigma1 = self.sigma1.map['values'][..., np.newaxis]
            sigma2 = self.sigma2.map['values'][..., np.newaxis]
            fraction = self.fraction.map['values'][..., np.newaxis]
            centre = self.centre.map['values'][..., np.newaxis]
        else:
            x = axis
            A = self.A.value
            sigma1 = self.sigma1.value
            sigma2 = self.sigma2.value
            fraction = self.fraction.value
            centre = self.centre.value
        return self._function(x, A, sigma1, sigma2, fraction, centre)

    function_nd.__doc__ %= FUNCTION_ND_DOCSTRING

    def estimate_parameters(self, signal, x1, x2, only_current=False):
        """Estimate the split voigt function by calculating the
           momenta the gaussian.

        Parameters
        ----------
        signal : Signal1D instance
        x1 : float
            Defines the left limit of the spectral range to use for the
            estimation.
        x2 : float
            Defines the right limit of the spectral range to use for the
            estimation.

        only_current : bool
            If False estimates the parameters for the full dataset.

        Returns
        -------
        bool

        Notes
        -----
        Adapted from http://www.scipy.org/Cookbook/FittingData

        Examples
        --------

        >>> g = hs.model.components1D.Gaussian()
        >>> x = np.arange(-10,10, 0.01)
        >>> data = np.zeros((32,32,2000))
        >>> data[:] = g.function(x).reshape((1,1,2000))
        >>> s = hs.signals.Signal1D({'data' : data})
        >>> s.axes_manager.axes[-1].offset = -10
        >>> s.axes_manager.axes[-1].scale = 0.01
        >>> g.estimate_parameters(s, -10,10, False)

        """
        super(SplitVoigt, self)._estimate_parameters(signal)
        axis = signal.axes_manager.signal_axes[0]
        centre, height, sigma = _estimate_gaussian_parameters(signal, x1, x2,
                                                              only_current)

        if only_current is True:
            self.centre.value = centre
            self.sigma1.value = sigma
            self.sigma2.value = sigma
            self.A.value = height * sigma * sqrt2pi
            if self.binned:
                self.A.value /= axis.scale
            return True
        else:
            if self.A.map is None:
                self._create_arrays()
            self.A.map['values'][:] = height * sigma * sqrt2pi
            if self.binned:
                self.A.map['values'][:] /= axis.scale
            self.A.map['is_set'][:] = True
            self.sigma1.map['values'][:] = sigma
            self.sigma1.map['is_set'][:] = True
            self.sigma2.map['values'][:] = sigma
            self.sigma2.map['is_set'][:] = True
            self.centre.map['values'][:] = centre
            self.centre.map['is_set'][:] = True
            self.fetch_stored_values()
            return True

    @property
    def height(self):
        return self.A.value / (self.sigma.value * sqrt2pi)

    @height.setter
    def height(self, value):
        self.A.value = value * self.sigma.value * sqrt2pi



# Copyright 2007-2020 The HyperSpy developers
#
# This file is part of  HyperSpy.
#
#  HyperSpy is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
#  HyperSpy is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with  HyperSpy.  If not, see <http://www.gnu.org/licenses/>.


import numpy as np
import pytest
from pathlib import Path

from hyperspy import signals, datasets
from hyperspy._components.gaussian import Gaussian
from hyperspy.decorators import lazifyTestClass
from hyperspy.io import load

MYPATH = Path(__file__).resolve().parent


@lazifyTestClass
class Test_Estimate_Elastic_Scattering_Threshold:

    def setup_method(self, method):
        # Create an empty spectrum
        s = signals.EELSSpectrum(np.zeros((3, 2, 1024)))
        energy_axis = s.axes_manager.signal_axes[0]
        energy_axis.scale = 0.02
        energy_axis.offset = -5

        gauss = Gaussian()
        gauss.centre.value = 0
        gauss.A.value = 5000
        gauss.sigma.value = 0.5
        gauss2 = Gaussian()
        gauss2.sigma.value = 0.5
        # Inflexion point 1.5
        gauss2.A.value = 5000
        gauss2.centre.value = 5
        s.data[:] = (gauss.function(energy_axis.axis) +
                     gauss2.function(energy_axis.axis))
        self.signal = s

    def test_min_in_window_with_smoothing(self):
        s = self.signal
        thr = s.estimate_elastic_scattering_threshold(
            window=5,
            window_length=5,
            tol=0.00001,
        )
        np.testing.assert_allclose(thr.data, 2.5, atol=10e-3)
        assert thr.metadata.Signal.signal_type == ""
        assert thr.axes_manager.signal_dimension == 0

    def test_min_in_window_without_smoothing_single_spectrum(self):
        s = self.signal.inav[0, 0]
        thr = s.estimate_elastic_scattering_threshold(
            window=5,
            window_length=0,
            tol=0.001,
        )
        np.testing.assert_allclose(thr.data, 2.49, atol=10e-3)

    def test_min_in_window_without_smoothing(self):
        s = self.signal
        thr = s.estimate_elastic_scattering_threshold(
            window=5,
            window_length=0,
            tol=0.001,
        )
        np.testing.assert_allclose(thr.data, 2.49, atol=10e-3)

    def test_min_not_in_window(self):
        # If I use a much lower window, this is the value that has to be
        # returned as threshold.
        s = self.signal
        data = s.estimate_elastic_scattering_threshold(window=1.5,
                                                       tol=0.001,
                                                       ).data
        assert np.all(np.isnan(data))

    def test_estimate_elastic_scattering_intensity(self):
        s = self.signal
        threshold = s.estimate_elastic_scattering_threshold(window=4.)
        # Threshold is nd signal
        t = s.estimate_elastic_scattering_intensity(threshold=threshold)
        assert t.metadata.Signal.signal_type == ""
        assert t.axes_manager.signal_dimension == 0
        np.testing.assert_array_almost_equal(t.data, 249999.985133)
        # Threshold is signal, 1 spectrum
        s0 = s.inav[0]
        t0 = s0.estimate_elastic_scattering_threshold(window=4.)
        t = s0.estimate_elastic_scattering_intensity(threshold=t0)
        np.testing.assert_array_almost_equal(t.data, 249999.985133)
        # Threshold is value
        t = s.estimate_elastic_scattering_intensity(threshold=2.5)
        np.testing.assert_array_almost_equal(t.data, 249999.985133)


@lazifyTestClass
class TestEstimateZLPCentre:

    def setup_method(self, method):
        s = signals.EELSSpectrum(np.diag(np.arange(1, 11)))
        s.axes_manager[-1].scale = 0.1
        s.axes_manager[-1].offset = 100
        self.signal = s

    def test_estimate_zero_loss_peak_centre(self):
        s = self.signal
        zlpc = s.estimate_zero_loss_peak_centre()
        np.testing.assert_allclose(zlpc.data,
                                   np.arange(100,
                                             101,
                                             0.1))
        assert zlpc.metadata.Signal.signal_type == ""
        assert zlpc.axes_manager.signal_dimension == 0


@lazifyTestClass
class TestAlignZLP:

    def setup_method(self, method):
        s = signals.EELSSpectrum(np.zeros((10, 100)))
        self.scale = 0.1
        self.offset = -2
        eaxis = s.axes_manager.signal_axes[0]
        eaxis.scale = self.scale
        eaxis.offset = self.offset
        self.izlp = eaxis.value2index(0)
        self.bg = 2
        self.ishifts = np.array([0, 4, 2, -2, 5, -2, -5, -9, -9, -8])
        self.new_offset = self.offset - self.ishifts.min() * self.scale
        s.data[np.arange(10), self.ishifts + self.izlp] = 10
        s.data += self.bg
        s.axes_manager[-1].offset += 100
        self.signal = s

    def test_align_zero_loss_peak_calibrate_true(self):
        s = self.signal
        s.align_zero_loss_peak(
            calibrate=True,
            print_stats=False)
        zlpc = s.estimate_zero_loss_peak_centre()
        np.testing.assert_allclose(zlpc.data.mean(), 0)
        np.testing.assert_allclose(zlpc.data.std(), 0)

    def test_align_zero_loss_peak_calibrate_true_with_mask(self):
        s = self.signal
        mask = s._get_navigation_signal(dtype="bool").T
        mask.data[[3, 5]] = (True, True)
        s.align_zero_loss_peak(
            calibrate=True,
            print_stats=False,
            mask=mask)
        zlpc = s.estimate_zero_loss_peak_centre(mask=mask)
        np.testing.assert_allclose(np.nanmean(zlpc.data), 0,
                                   atol=np.finfo(float).eps)
        np.testing.assert_allclose(np.nanstd(zlpc.data), 0,
                                   atol=np.finfo(float).eps)

    def test_align_zero_loss_peak_calibrate_false(self):
        s = self.signal
        s.align_zero_loss_peak(
            calibrate=False,
            print_stats=False)
        zlpc = s.estimate_zero_loss_peak_centre()
        np.testing.assert_allclose(zlpc.data.std(), 0, atol=10e-3)

    def test_also_aligns(self):
        s = self.signal
        s2 = s.deepcopy()
        s.align_zero_loss_peak(calibrate=True,
                               print_stats=False,
                               also_align=[s2])
        zlpc = s2.estimate_zero_loss_peak_centre()
        assert zlpc.data.mean() == 0
        assert zlpc.data.std() == 0

    def test_align_zero_loss_peak_with_spike_signal_range(self):
        s = self.signal
        spike = np.zeros((10, 100))
        spike_amplitude = 20
        spike[:, 75] = spike_amplitude
        s.data += spike
        s.align_zero_loss_peak(
            print_stats=False, subpixel=False, signal_range=(98., 102.))
        zlp_max = s.isig[-0.5:0.5].max(-1).data
        # Max value in the original spectrum is 12, but due to the aligning
        # the peak is split between two different channels. So 8 is the
        # maximum value for the aligned spectrum
        np.testing.assert_allclose(zlp_max, 8)

    def test_align_zero_loss_peak_crop_false(self):
        s = self.signal
        original_size = s.axes_manager.signal_axes[0].size
        s.align_zero_loss_peak(
            crop=False,
            print_stats=False)
        assert original_size == s.axes_manager.signal_axes[0].size


@lazifyTestClass
class TestPowerLawExtrapolation:

    def setup_method(self, method):
        s = signals.EELSSpectrum(0.1 * np.arange(50, 250, 0.5) ** -3.)
        s.metadata.Signal.binned = False
        s.axes_manager[-1].offset = 50
        s.axes_manager[-1].scale = 0.5
        self.s = s

    def test_unbinned(self):
        sc = self.s.isig[:300]
        s = sc.power_law_extrapolation(extrapolation_size=100)
        np.testing.assert_allclose(s.data, self.s.data, atol=10e-3)

    def test_binned(self):
        self.s.data *= self.s.axes_manager[-1].scale
        self.s.metadata.Signal.binned = True
        sc = self.s.isig[:300]
        s = sc.power_law_extrapolation(extrapolation_size=100)
        np.testing.assert_allclose(s.data, self.s.data, atol=10e-3)


@lazifyTestClass
class TestFourierRatioDeconvolution:

    @pytest.mark.parametrize(('extrapolate_lowloss'), [True, False])
    def test_running(self, extrapolate_lowloss):
        s = signals.EELSSpectrum(np.arange(200))
        gaussian = Gaussian()
        gaussian.A.value = 50
        gaussian.sigma.value = 10
        gaussian.centre.value = 20
        s_ll = signals.EELSSpectrum(gaussian.function(np.arange(0, 200, 1)))
        s_ll.axes_manager[0].offset = -50
        s.fourier_ratio_deconvolution(s_ll,
                                      extrapolate_lowloss=extrapolate_lowloss)


class TestRebin:
    def setup_method(self, method):
        # Create an empty spectrum
        s = signals.EELSSpectrum(np.ones((4, 2, 1024)))
        self.signal = s

    def test_rebin_without_dwell_time(self):
        s = self.signal
        s.rebin(scale=(2, 2, 1))

    def test_rebin_dwell_time(self):
        s = self.signal
        s.metadata.add_node("Acquisition_instrument.TEM.Detector.EELS")
        s_mdEELS = s.metadata.Acquisition_instrument.TEM.Detector.EELS
        s_mdEELS.dwell_time = 0.1
        s_mdEELS.exposure = 0.5
        s2 = s.rebin(scale=(2, 2, 8))
        s2_mdEELS = s2.metadata.Acquisition_instrument.TEM.Detector.EELS
        assert s2_mdEELS.dwell_time == (0.1 * 2 * 2)
        assert s2_mdEELS.exposure == (0.5 * 2 * 2)

        def test_rebin_exposure(self):
            s = self.signal
            s.metadata.exposure = 10.2
            s2 = s.rebin(scale=(2, 2, 8))
            assert s2.metadata.exposure == (10.2 * 2 * 2)

    def test_offset_after_rebin(self):
        s = self.signal
        s.axes_manager[0].offset = 1
        s.axes_manager[1].offset = 2
        s.axes_manager[2].offset = 3
        s2 = s.rebin(scale=(2, 2, 1))
        assert s2.axes_manager[0].offset == 1.5
        assert s2.axes_manager[1].offset == 2.5
        assert s2.axes_manager[2].offset == s.axes_manager[2].offset

@lazifyTestClass
class Test_Estimate_Thickness:

    def setup_method(self, method):
        # Create an empty spectrum
        self.s = load(
            MYPATH.joinpath("data/EELS_LL_linescan_simulated_thickness_variation.hspy")
        )
        self.zlp = load(
            MYPATH.joinpath("data/EELS_ZLP_linescan_simulated_thickness_variation.hspy")
        )

    def test_relative_thickness(self):
        t = self.s.estimate_thickness(zlp=self.zlp)
        np.testing.assert_allclose(t.data, np.arange(0.3,2,0.1), atol=4e-3)
        assert t.metadata.Signal.quantity == "$\\frac{t}{\\lambda}$"

    def test_thickness_mfp(self):
        t = self.s.estimate_thickness(zlp=self.zlp, mean_free_path=120)
        np.testing.assert_allclose(t.data, 120 * np.arange(0.3,2,0.1), rtol=3e-3)
        assert t.metadata.Signal.quantity == "thickness (nm)"

    def test_thickness_density(self):
        t = self.s.estimate_thickness(zlp=self.zlp, density=3.6)
        np.testing.assert_allclose(t.data, 142 * np.arange(0.3,2,0.1), rtol=3e-3)
        assert t.metadata.Signal.quantity == "thickness (nm)"

    def test_thickness_density_and_mfp(self):
        t = self.s.estimate_thickness(zlp=self.zlp, density=3.6, mean_free_path=120)
        np.testing.assert_allclose(t.data, 127.5 * np.arange(0.3,2,0.1), rtol=3e-3)
        assert t.metadata.Signal.quantity == "thickness (nm)"

    def test_threshold(self):
        t = self.s.estimate_thickness(threshold=4.5, density=3.6, mean_free_path=120)
        np.testing.assert_allclose(t.data, 127.5 * np.arange(0.3,2,0.1), rtol=3e-3)
        assert t.metadata.Signal.quantity == "thickness (nm)"

    def test_threshold_nd(self):
        threshold = self.s._get_navigation_signal()
        threshold.data[:] = 4.5
        t = self.s.estimate_thickness(threshold=threshold, density=3.6, mean_free_path=120)
        np.testing.assert_allclose(t.data, 127.5 * np.arange(0.3,2,0.1), rtol=3e-3)
        assert t.metadata.Signal.quantity == "thickness (nm)"

    def test_no_zlp_or_threshold(self):
        with pytest.raises(ValueError):
            self.s.estimate_thickness()

    def test_no_metadata(self):
        del self.s.metadata.Acquisition_instrument
        with pytest.raises(RuntimeError):
            self.s.estimate_thickness(zlp=self.zlp, density=3.6)

class Test_Print_Edges_Near_Energy:
    def setup_method(self, method):
        # Create an empty spectrum
        s = signals.EELSSpectrum(np.ones((4, 2, 1024)))
        self.signal = s

    def test_at_532eV(self):
        s = self.signal
        table_ascii = s.print_edges_near_energy(532)

        assert table_ascii.__repr__() == ('+-------+-------------------+------'
        '-----+-----------------+\n|  edge | onset energy (eV) | relevance '
        '|   description   |\n+-------+-------------------+-----------+'
        '-----------------+\n|  O_K  |       532.0       |   Major   '
        '|   Abrupt onset  |\n| Pd_M3 |       531.0       |   Minor   '
        '|                 |\n| Sb_M5 |       528.0       |   Major   '
        '| Delayed maximum |\n| Sb_M4 |       537.0       |   Major   '
        '| Delayed maximum |\n+-------+-------------------+-----------+'
        '-----------------+')

    def test_sequence_edges(self):
        s = self.signal
        table_ascii = s.print_edges_near_energy(123,
                                                edges=['Mn_L2', 'O_K', 'Fe_L2'])

        assert table_ascii.__repr__() == ('+-------+-------------------+------'
        '-----+-----------------------------+\n|  edge | onset energy (eV) '
        '| relevance |         description         |\n+-------+------------'
        '-------+-----------+-----------------------------+\n| Mn_L2 |     '
        '  651.0       |   Major   | Sharp peak. Delayed maximum |\n|  O_K  |'
        '       532.0       |   Major   |         Abrupt onset        '
        '|\n| Fe_L2 |       721.0       |   Major   | Sharp peak. '
        'Delayed maximum |\n+-------+-------------------+-----------+---------'
        '--------------------+')

    def test_no_energy_and_edges(self):
        s = self.signal
        with pytest.raises(ValueError):
            s.print_edges_near_energy()

class Test_Edges_At_Energy:
    def setup_method(self, method):
        # Create an empty spectrum
        s = signals.EELSSpectrum(np.ones((4, 2, 1024)))
        self.signal = s

    def test_at_532eV(self):
        s = self.signal
        table_ascii = s.edges_at_energy(532, width=20, only_major=True,
                                        order='ascending')

        assert table_ascii.__repr__() == ('+-------+-------------------+------'
        '-----+-----------------+\n|  edge | onset energy (eV) | relevance |  '
        ' description   |\n+-------+-------------------+-----------+----------'
        '-------+\n| Sb_M5 |       528.0       |   Major   | Delayed maximum '
        '|\n|  O_K  |       532.0       |   Major   |   Abrupt onset  |\n| '
        'Sb_M4 |       537.0       |   Major   | Delayed maximum |\n+-------+'
        '-------------------+-----------+-----------------+')

class Test_Get_Complementary_Edges:
    def setup_method(self, method):
        # Create an empty spectrum
        s = signals.EELSSpectrum(np.ones((4, 2, 1024)))
        self.signal = s

    def test_Fe_O(self):
        s = self.signal
        complementary = s.get_complementary_edges(['Fe_L2', 'O_K'])

        assert complementary == ['Fe_L1', 'Fe_L3', 'Fe_M3', 'Fe_M2']

    def test_Fe_O_only_major(self):
        s = self.signal
        complementary = s.get_complementary_edges(['Fe_L2', 'O_K'],
                                                  only_major=True)

        assert complementary == ['Fe_L3', 'Fe_M3', 'Fe_M2']

class Test_Plot_EELS:
    def setup_method(self, method):
        s = datasets.artificial_data.get_core_loss_eels_signal()
        self.signal = s

    def test_plot_no_markers(self):
        s = self.signal
        s.add_elements(('Mn','Cr'))
        s.plot()

        assert len(s._edge_markers) == 0

    def test_plot_edges_True(self):
        s = self.signal
        s.add_elements(('Mn','Cr'))
        s.plot(plot_edges=True)

        assert len(s._edge_markers) == 6
        assert set(s._edge_markers.keys()) == set(['Mn_L2', 'Mn_L3', 'Mn_L1',
                                                   'Cr_L2', 'Cr_L3', 'Cr_L1'])

    def test_plot_edges_True_without_elements(self):
        s = self.signal
        with pytest.raises(ValueError):
            s.plot(plot_edges=True)

    def test_plot_edges_from_element_family_specific(self):
        s = self.signal
        s.plot(plot_edges=['Mn', 'Ti_L', 'Cr_L3'], only_edges=('Major'))

        assert len(s._edge_markers) == 5
        assert set(s._edge_markers.keys()) == set(['Cr_L3', 'Mn_L2', 'Mn_L3',
                                                   'Ti_L2', 'Ti_L3'])

    def test_unsupported_edge_family(self):
        s = self.signal
        with pytest.raises(AttributeError):
            s.plot(plot_edges=['Cr_P'])

    def test_unsupported_edge(self):
        s = self.signal
        with pytest.raises(AttributeError):
            s.plot(plot_edges=['Xe_P4'])

    def test_unsupported_element(self):
        s = self.signal
        with pytest.raises(ValueError):
            s.plot(plot_edges=['ABC_L1'])

    def test_remove_edge_labels(self):
        s = self.signal
        s.plot(plot_edges=['Cr_L', 'Fe_L2'])
        s.remove_EELS_edges_markers(['Cr_L1', 'Fe_L2'])

        assert len(s._edge_markers) == 2
        assert set(s._edge_markers.keys()) == set(['Cr_L2', 'Cr_L3'])

    def test_unequal_edges_and_markers(self):
        s = self.signal
        s.plot()
        with pytest.raises(ValueError):
            s.plot_edges_label(['Cr_L3', 'Fe_L2'],
                               vertical_line_marker=['vl1', 'vl2'],
                               text_marker=['tx1'])

    def test_plot_edges_without_markers_provided(self):
        s = self.signal
        s.plot()
        s.plot_edges_label({'Fe_L2': 721.0, 'O_K': 532.0})

        assert len(s._edge_markers) == 2
        assert set(s._edge_markers.keys()) == set(['Fe_L2', 'O_K'])
